# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import TypeVar, Tuple, Sequence, Callable, Optional, Generic
from numpy import ndarray, double
from .config_types import Setting

FVal = TypeVar('FVal')

class ObjFunc(Generic[FVal]):

    @abstractmethod
    def fitness(self, v: ndarray) -> double:
        """(`cdef` function) Return the fitness from the variable list `v`.
        This function will be directly called in the algorithms.
        """
        ...

    def result(self, v: ndarray) -> FVal:
        ...

class Algorithm(Generic[FVal]):
    func: ObjFunc[FVal]

    def __class_getitem__(cls, item):
        # PEP 560
        raise NotImplemented

    @abstractmethod
    def __init__(
        self,
        func: ObjFunc[FVal],
        settings: Setting,
        progress_fun: Optional[Callable[[int, str], None]] = None,
        interrupt_fun: Optional[Callable[[], bool]] = None
    ):
        """The argument `func` is an object inherit from [ObjFunc],
        and all abstract methods should be implemented.

        The format of argument `settings` can be customized.

        The argument `progress_fun` will be called when update progress,
        and the argument `interrupt_fun` will check the interrupt status from
        GUI or subprocess.
        """
        ...

    def history(self) -> ndarray:
        ...

    def result(self) -> Tuple[ndarray, float]:
        ...

    def run(self) -> FVal:
        ...



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Test objective function for algorithms.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from numpy import array, float64 as f64
from .utility cimport ObjFunc


@cython.final
cdef class TestObj(ObjFunc):
    """Test objective function.

    f(x) = x1^2 + 8*x2
    """

    def __cinit__(self):
        self.ub = array([50] * 4, dtype=f64)
        self.lb = array([0] * 4, dtype=f64)

    cdef double target(self, double[:] v) nogil:
        return v[0] * v[0] + 8 * v[1] * v[1] + v[2] * v[2] + v[3] * v[3]

    cdef double fitness(self, double[:] v) nogil:
        return self.target(v)

    cpdef object result(self, double[:] v):
        return self.target(v)



# -*- coding: utf-8 -*-

"""Pyslvs optimization targets."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from .f_planar import (
    FPlanar, norm_path, curvature, derivative, path_signature,
    cross_correlation,
)
from .n_planar import NPlanar, norm_pca
from .utility import FConfig, NConfig



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Planarity check function for Graph class.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from .graph cimport Graph

cpdef bint is_planar(Graph g)



# -*- coding: utf-8 -*-

from typing import Tuple, Iterable
from numpy import ndarray, double
from pyslvs.metaheuristics import ObjFunc
from .utility import FConfig

_Coord = Tuple[float, float]

def norm_path(path: Iterable[_Coord], scale: float = 1) -> ndarray:
    ...

def curvature(path: Iterable[_Coord]) -> ndarray:
    ...

def derivative(path: ndarray) -> ndarray:
    ...

def path_signature(k: ndarray, maximum: float = 100) -> ndarray:
    ...

def cross_correlation(p1: ndarray, p2: ndarray, t: float = 0.1) -> ndarray:
    ...


class FPlanar(ObjFunc[str]):
    callback: int

    def __init__(self, mech: FConfig):
        """The constructor of objective object.

        Options:
        + `expression`: The mechanism expression of the structure.
        + `input`: Input pairs.
        + `placement`: The grounded joints setting. (`x`, `y`, `r`)
        + `target`: The target path.
        + `same`: Multiple joint setting. The joints are according to [`edges_view`](#edges_view).
        + `upper`: The upper setting of variables, the length must same as variable array.
        + `lower`: The lower setting of variables, the length must same as variable array.
        + `shape_only`: Compare paths by shape only.

        Variable array:

        | | Placement | Link length | Inputs |
        |:---:|:-----:|:-----------:|:------:|
        | `v =` | `x0`, `y0`, ... | `l0`, `l1`, ... | `a00`, `a01`, ..., `a10`, `a11`, ... |

        In 1D array: `v = [x0, y0, ..., l0, l1, ..., a00, a01, ..., a10, a11, ...]`
        """
        ...

    def fitness(self, v: ndarray) -> double:
        """The fitness is the error between target path and self.

        Chromosome format: (decided by upper and lower)

        v: `[Ax, Ay, Dx, Dy, ..., L0, L1, ..., A00, A01, ..., A10, A11, ...]`
        """
        ...

    def is_two_kernel(self) -> bool:
        ...

    def result(self, v: ndarray) -> str:
        ...



# -*- coding: utf-8 -*-

from typing import Tuple, List, Sequence, Mapping, Union, Optional
from numpy import ndarray
from .topo_config import EStack
from .expression import VPoint, Coord

_Coord = Tuple[float, float]

def pxy(c1: Coord, x: float, y: float) -> Coord:
    ...

def plap(
    c1: Coord,
    d0: float,
    a0: float,
    c2: Optional[Coord] = None,
    inverse: bool = False
) -> Coord:
    ...

def ppp(c1: Coord, c2: Coord, c3: Coord) -> Coord:
    ...

def pllp(c1: Coord, d0: float, d1: float, c2: Coord,
         inverse: bool = False) -> Coord:
    ...

def plpp(c1: Coord, d0: float, c2: Coord, c3: Coord,
         inverse: bool = False) -> Coord:
    ...

def palp(c1: Coord, a0: float, d0: float, c2: Coord, inverse: bool = False) -> Coord:
    ...

def vpoint_dof(vpoints: Sequence[VPoint]) -> int:
    ...

def expr_solving(
    exprs: EStack,
    vpoints: Sequence[VPoint],
    angles: Optional[Mapping[Tuple[int, int], float]] = None
) -> List[Union[_Coord, Tuple[_Coord, _Coord]]]:
    ...

def uniform_four_bar(ml: float, n: int) -> ndarray:
    ...

def uniform_path(v: ndarray, n: int) -> ndarray:
    ...

def uniform_expr(v: ndarray) -> List[VPoint]:
    ...



# -*- coding: utf-8 -*-

"""Pyslvs core module test."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from math import sqrt, radians
from pyslvs import (
    Coord, SolverSystem, pxy, ppp, plap, pllp, plpp, palp, expr_solving,
    t_config, parse_vpoints, example_list,
)
from . import TestBase


class CoreTest(TestBase):

    def test_pxy(self):
        """Test for pxy function."""
        coord = pxy(Coord(80, 90), 40, -20)
        self.assertAlmostEqual(120, coord.x)
        self.assertAlmostEqual(70, coord.y)

    def test_ppp(self):
        """Test for ppp function."""
        coord = ppp(Coord(0, 0), Coord(0, 90), Coord(90, 0))
        self.assertAlmostEqual(90, coord.x)
        self.assertAlmostEqual(90, coord.y)

    def test_plap(self):
        """Test for plap function."""
        coord = plap(Coord(0, 0), 50 * sqrt(2), radians(45), Coord(50, 0))
        self.assertAlmostEqual(50, coord.x)
        self.assertAlmostEqual(50, coord.y)

    def test_pllp(self):
        """Test for pllp function."""
        c1 = Coord(-30, 0)
        c2 = Coord(30, 0)
        coord = pllp(c1, 50, 50, c2)
        self.assertAlmostEqual(0, coord.x)
        self.assertAlmostEqual(40, coord.y)
        coord = pllp(c1, 30, 30, c2)
        self.assertAlmostEqual(coord.x, 0)
        self.assertAlmostEqual(coord.y, 0)
        coord = pllp(c1, 90, 30, c2)
        self.assertAlmostEqual(60, coord.x)
        self.assertAlmostEqual(0, coord.y)

    def test_plpp(self):
        """Test for plpp function."""
        coord = plpp(Coord(0, 0), sqrt(5), Coord(0, -3), Coord(3 / 2, 0))
        self.assertAlmostEqual(2, coord.x)
        self.assertAlmostEqual(1, coord.y)

    def test_palp(self):
        """Test for palp function."""
        coord = palp(Coord(0, 0), radians(15), 20, Coord(60, 10))
        self.assertAlmostEqual(42.253221, coord.x, 6)
        self.assertAlmostEqual(19.222356, coord.y, 6)

    def test_solving(self):
        """Test triangular formula solving.

        + Test for PMKS parser.
        + Test data collecting function.
        + Test expression solving function.
        """

        def test_case(name: str):
            expr, inputs = example_list(name)
            vpoints = parse_vpoints(expr)
            exprs = t_config(vpoints, inputs)
            result = expr_solving(exprs, vpoints, {pair: 0. for pair in inputs})
            return result[-1]

        x, y = test_case("Jansen's linkage (Single)")
        self.assertAlmostEqual(-43.170055, x, 6)
        self.assertAlmostEqual(-91.753226, y, 6)
        x, y = test_case("Crank slider (RP joint)")
        self.assertAlmostEqual(103.801126, x, 6)
        self.assertAlmostEqual(78.393173, y, 6)
        x, y = test_case("Parallel Linkage")
        self.assertAlmostEqual(200, x, 6)
        self.assertAlmostEqual(0, y, 6)
        # TODO: New test case for Inverted slider

    def test_solving_bfgs(self):
        """Test Sketch Solve kernel."""
        expr, _ = example_list("Jansen's linkage (Single)")
        system = SolverSystem(parse_vpoints(expr), {(0, 1): 0.})
        result = system.solve()
        x, y = result[7]
        self.assertAlmostEqual(-43.170055, x, 6)
        self.assertAlmostEqual(-91.753226, y, 6)
        # Test if angle value changed
        system.set_inputs({(0, 1): 45.})
        result = system.solve()
        x, y = result[7]
        self.assertAlmostEqual(-24.406394, x, 6)
        self.assertAlmostEqual(-91.789596, y, 6)
        # Test if link length changed
        system.set_data({(0, 1): 16.})
        result = system.solve()
        x, y = result[7]
        self.assertAlmostEqual(-24.117994, x, 6)
        self.assertAlmostEqual(-91.198072, y, 6)



# -*- coding: utf-8 -*-

"""Kernel of Pyslvs."""

__version__ = "21.11.0"
__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from .expression import get_vlinks, VJoint, VPoint, VLink, Coord
from .bfgs import SolverSystem
from .topo_config import t_config, EStack
from .tinycadlib import (
    pxy, ppp, plap, pllp, plpp, palp, vpoint_dof, expr_solving, uniform_path,
    uniform_four_bar, uniform_expr,
)
from .expression_parser import (
    color_names, color_rgb, parse_params, parse_pos, parse_vpoints,
    parse_vlinks, edges_view, graph2vpoints, PointArgs, LinkArgs,
)
from .example import example_list, all_examples
from .collection import collection_list, all_collections, Collection
from .efd import efd_fitting


def get_include() -> str:
    """Get include directory."""
    from os.path import dirname
    return dirname(__file__)



# -*- coding: utf-8 -*-

"""The source code refer from "spatial_efd" module."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from typing import Tuple, Sized, Sequence, Union, Optional
from numpy import (
    pi, sqrt, abs, cos, sin, arctan2, array, ndarray, linspace, zeros, ones,
    asarray, diff, concatenate, cumsum,
)

_Path = Union[Sequence[Tuple[float, float]], ndarray]


def efd_fitting(path: _Path, n: int = 0,
                harmonic: Optional[int] = None) -> ndarray:
    """Curve fitting using Elliptical Fourier Descriptor.

    The path `path` will be translated to Fourier descriptor coefficients,
    then regenerate a new path as a `n` x 4 NumPy array.
    """
    contour = asarray(path, dtype=float)
    if n < 3:
        n = len(contour)
    if harmonic is None:
        harmonic = fourier_power(
            calculate_efd(contour, _nyquist(contour)),
            _nyquist(contour)
        )
    coeffs, rot = normalize_efd(calculate_efd(contour, harmonic), norm=False)
    locus_v = locus(contour)
    # New path
    contour = inverse_transform(coeffs, locus_v, n, harmonic)
    return rotate_contour(contour, -rot, locus_v)


def normalize_efd(
    coeffs: ndarray,
    norm: bool = True
) -> Tuple[ndarray, float]:
    """
    Normalize the Elliptical Fourier Descriptor coefficients for a polygon.

    Implements Kuhl and Giardina method of normalizing the coefficients
    An, Bn, Cn, Dn. Performs 3 separate normalizations. First, it makes the
    data location invariant by re-scaling the data to a common origin.
    Secondly, the data is rotated with respect to the major axis. Thirdly,
    the coefficients are normalized with regard to the absolute value of A_1.
    This code is adapted from the pyefd module. See the original paper for
    more detail:

    Kuhl, FP and Giardina, CR (1982). Elliptic Fourier features of a closed
    contour. Computer graphics and image procesnp_sing, 18(3), 236-258.

    Args:
        coeffs: A numpy array of shape (n, 4) representing the
            four coefficients for each harmonic computed.
        norm: Set to True (the default) to perform the third
            normalization and false to return the data withot this procesnp_sing
            step. Set this to False when plotting a comparison between the
            input data and the Fourier ellipse.
    Returns:
        A tuple consisting of a numpy.ndarray of shape (harmonics, 4)
        representing the four coefficients for each harmonic computed and
        the rotation in radians applied to the normalized contour.
    """
    # Make the coefficients have a zero phase shift from
    # the first major axis. Theta_1 is that shift angle.
    theta1 = 0.5 * arctan2(
        2 * (coeffs[0, 0] * coeffs[0, 1] + coeffs[0, 2] * coeffs[0, 3]),
        coeffs[0, 0] ** 2 - coeffs[0, 1] ** 2
        + coeffs[0, 2] ** 2 - coeffs[0, 3] ** 2
    )
    coeffs = coeffs.copy()
    # Rotate all coefficients by theta_1
    for n in range(coeffs.shape[0]):
        angle = (n + 1) * theta1
        coeffs[n, :] = (array([
            [coeffs[n, 0], coeffs[n, 1]],
            [coeffs[n, 2], coeffs[n, 3]],
        ]) @ array([
            [cos(angle), -sin(angle)],
            [sin(angle), cos(angle)],
        ])).flat
    # Make the coefficients rotation invariant by rotating so that
    # the semi-major axis is parallel to the x-axis.
    psi1 = arctan2(coeffs[0, 2], coeffs[0, 0])
    psi2 = array([
        [cos(psi1), sin(psi1)],
        [-sin(psi1), cos(psi1)],
    ])
    # Rotate all coefficients by -psi_1.
    for n in range(coeffs.shape[0]):
        coeffs[n, :] = (psi2 @ array([
            [coeffs[n, 0], coeffs[n, 1]],
            [coeffs[n, 2], coeffs[n, 3]],
        ])).flat
    if norm:
        # Obtain size-invariance by normalizing.
        coeffs /= abs(coeffs[0, 0])
    return coeffs, psi1


def locus(contour: ndarray) -> Tuple[float, float]:
    """
    Compute the dc coefficients, used as the locus when calling
    inverse_transform().

    This code is adapted from the pyefd module. See the original paper for
    more detail:

    Kuhl, FP and Giardina, CR (1982). Elliptic Fourier features of a closed
    contour. Computer graphics and image procesnp_sing, 18(3), 236-258.

    Args:
        contour: A n x 2 numpy array represents a path.
    Returns:
        A tuple containing the c and d coefficients.
    """
    dxy = diff(contour, axis=0)
    dt = sqrt((dxy ** 2).sum(axis=1))
    t = concatenate(([0], cumsum(dt)))
    zt = t[-1]
    xi = cumsum(dxy[:, 0]) - dxy[:, 0] / dt * t[1:]
    c = diff(t ** 2) / (dt * 2)
    a0 = sum(dxy[:, 0] * c + xi * dt) / (zt + 1e-20)
    delta = cumsum(dxy[:, 1]) - dxy[:, 1] / dt * t[1:]
    c0 = sum(dxy[:, 1] * c + delta * dt) / (zt + 1e-20)
    # A0 and CO relate to the first point of the contour array as origin
    # Adding those values to the coefficients to make them relate to true origin
    return contour[0, 0] + a0, contour[0, 1] + c0


def calculate_efd(contour: ndarray, harmonic: int = 10) -> ndarray:
    """
    Compute the Elliptical Fourier Descriptors for a polygon.

    Implements Kuhl and Giardina method of computing the coefficients
    An, Bn, Cn, Dn for a specified number of harmonics. This code is adapted
    from the pyefd module. See the original paper for more detail:

    Kuhl, FP and Giardina, CR (1982). Elliptic Fourier features of a closed
    contour. Computer graphics and image processing, 18(3), 236-258.

    Args:
        contour: A n x 2 numpy array represents a path.
        harmonic: The number of harmonics to compute for the given
            shape, defaults to 10.
    Returns:
        A numpy array of shape (harmonics, 4) representing the
        four coefficients for each harmonic computed.
    """
    dxy = diff(contour, axis=0)
    dt = sqrt((dxy ** 2).sum(axis=1))
    t = concatenate(([0], cumsum(dt)))
    zt = t[-1]
    phi = (2. * pi * t) / (zt + 1e-20)
    coeffs = zeros((harmonic, 4))
    for n in range(1, harmonic + 1):
        c = zt / (2 * n * n * pi * pi)
        phi_n = phi * n
        cos_phi_n = (cos(phi_n[1:]) - cos(phi_n[:-1])) / dt
        sin_phi_n = (sin(phi_n[1:]) - sin(phi_n[:-1])) / dt
        coeffs[n - 1, :] = (
            c * (dxy[:, 1] * cos_phi_n).sum(),
            c * (dxy[:, 1] * sin_phi_n).sum(),
            c * (dxy[:, 0] * cos_phi_n).sum(),
            c * (dxy[:, 0] * sin_phi_n).sum(),
        )
    return coeffs


def inverse_transform(
    coeffs: ndarray,
    locus_v: Tuple[float, float],
    n: int,
    harmonic: int
) -> ndarray:
    """
    Perform an inverse fourier transform to convert the coefficients back into
    spatial coordinates.

    Implements Kuhl and Giardina method of computing the performing the
    transform for a specified number of harmonics. This code is adapted
    from the pyefd module. See the original paper for more detail:

    Kuhl, FP and Giardina, CR (1982). Elliptic Fourier features of a closed
    contour. Computer graphics and image procesnp_sing, 18(3), 236-258.

    Args:
        coeffs: A numpy array of shape (n, 4) representing the
            four coefficients for each harmonic computed.
        locus_v: The x,y coordinates of the centroid of the contour being
            generated. Use calculate_dc_coefficients() to generate the correct
            locus for a shape.
        n: The number of coordinate pairs to compute. A larger value will
            result in a more complex shape at the expense of increased
            computational time.
        harmonic: The number of harmonics to be used to generate
            coordinates. Must be <= coeffs.shape[0]. Supply a
            smaller value to produce coordinates for a more generalized shape.
    Returns:
        A n x 2 numpy array represents a contour.
    """
    t = linspace(0, 1, n)
    contour = ones((n, 2), dtype=float)
    contour[:, 0] *= locus_v[0]
    contour[:, 1] *= locus_v[1]
    for n in range(harmonic):
        angle = t * (n + 1) * 2 * pi
        cosine = cos(angle)
        sine = sin(angle)
        contour[:, 0] += coeffs[n, 2] * cosine + coeffs[n, 3] * sine
        contour[:, 1] += coeffs[n, 0] * cosine + coeffs[n, 1] * sine
    return contour


def _nyquist(zx: Sized) -> int:
    """
    Returns the maximum number of harmonics that can be computed for a given
    contour, the Nyquist Frequency.

    See this paper for details:
    C. np_costa et al. / Postharvest Biology and Technology 54 (2009) 38-47

    Args:
        zx (list): A list (or numpy array) of x coordinate values.
    Returns:
        int: The nyquist frequency, expressed as a number of harmonics.
    """
    return len(zx) // 2


def fourier_power(
    coeffs: ndarray,
    nyq: int,
    threshold: float = 1.
) -> int:
    """
    Compute the total Fourier power and find the minimum number of harmonics
    required to exceed the threshold fraction of the total power.

    This is a good method for identifying the number of harmonics to use to
    describe a polygon. For more details see:

    C. np_costa et al. / Postharvest Biology and Technology 54 (2009) 38-47

    Warning:
        The number of coefficients must be >= the Nyquist Frequency.
    Args:
        coeffs: A numpy array of shape (n, 4) representing the
            four coefficients for each harmonic computed.
        nyq: The Nyquist Frequency.
        threshold: The threshold fraction of the total Fourier power,
            the default is 1.
    Returns:
        The number of harmonics required to represent the contour above
        the threshold Fourier power.
    """
    total_power = 0
    current_power = 0
    for i in range(nyq):
        total_power += 0.5 * (coeffs[i, :] ** 2).sum()
    for i in range(nyq):
        current_power += 0.5 * (coeffs[i, :] ** 2).sum()
        if current_power / total_power >= threshold:
            return i + 1
    return nyq


def rotate_contour(
    contour: ndarray,
    angle: float,
    centroid: Tuple[float, float]
) -> ndarray:
    """
    Rotates a contour about a point by a given amount expressed in degrees.

    Operates by calling rotatePoint() on each x,y pair in turn. X and Y must
    have the same dimensions.

    Args:
        contour: A n x 2 numpy array represents a path.
        angle: The angle in radians for the contour to be rotated by.
        centroid: A tuple containing the x,y coordinates of the centroid to
            rotate the contour about.
    Returns:
        A n x 2 numpy array represents a contour.
    """
    cpx, cpy = centroid
    out = zeros(contour.shape, dtype=contour.dtype)
    for i in range(len(contour)):
        dx = contour[i, 0] - cpx
        dy = contour[i, 1] - cpy
        out[i, :] = (
            cpx + dx * cos(angle) - dy * sin(angle),
            cpy + dx * sin(angle) + dy * cos(angle)
        )
    return out



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Tiny CAD library of PMKS symbolic and position analysis.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from libc.math cimport M_PI, NAN, sin, cos
from .expression cimport Coord, VJoint, VPoint, VLink, distance, slope_angle
from .bfgs cimport SolverSystem
from numpy import zeros, array
from numpy.random import uniform


def pxy(Coord c1, double x, double y):
    """The PXY function requires one point and offset values, get the
    position of second point.

    In the following picture, `c1` correspond to "A", `d0` correspond to "X",
    `d1` correspond to "Y", `return` correspond to "B", the sign of value are
    correspond to coordinate system.

    ![pxy](img/pxy.png)
    """
    cdef CCoord c = cpxy(CCoord(c1.x, c1.y), x, y)
    return Coord.__new__(Coord, c.x, c.y)


def ppp(Coord c1, Coord c2, Coord c3):
    """The PPP function is used to solve parallel linkage.

    In the following picture, `c1` correspond to "A", `c2` correspond to "B",
    `c3` correspond to "C", `return` correspond to "D".

    ![ppp](img/ppp.png)
    """
    cdef CCoord c = cppp(CCoord(c1.x, c1.y), CCoord(c2.x, c2.y),
                         CCoord(c3.x, c3.y))
    return Coord.__new__(Coord, c.x, c.y)


def plap(
    Coord c1,
    double d0,
    double a0,
    Coord c2 = None,
    bint inverse = False
):
    """The PLAP function requires two points, one distance and one angle,
    obtained the position of third point. The unit of `a0` is degree.

    In the following picture, `c1` correspond to "A", `c2` correspond to "B",
    `d0` correspond to "L0", `a0` correspond to "beta", `return` correspond
    to "C".
    If `c2` is not given, "alpha" will be set to zero.

    ![plap](img/plap.png)

    Set `inverse` option to `True` can make `a0` value as negative.
    """
    if c2 is None:
        c2 = c1
    cdef CCoord c = cplap(CCoord(c1.x, c1.y), d0, a0, CCoord(c2.x, c2.y), inverse)
    return Coord.__new__(Coord, c.x, c.y)


def pllp(
    Coord c1,
    double d0,
    double d1,
    Coord c2,
    bint inverse = False
):
    """The PLLP function requires two points and two distances, obtained the
    position of third point.

    In the following picture, `c1` correspond to "A", `c2` correspond to "B",
    `d0` correspond to "L0", `d1` correspond to "L1", `return` correspond to
    "C".

    ![pllp](img/pllp.png)

    Set `inverse` option to `True` can make the result upside down.
    """
    cdef CCoord c = cpllp(CCoord(c1.x, c1.y), d0, d1, CCoord(c2.x, c2.y), inverse)
    return Coord.__new__(Coord, c.x, c.y)


def plpp(
    Coord c1,
    double d0,
    Coord c2,
    Coord c3,
    bint inverse = False
):
    """The PLPP function requires three points and one distance, obtained the
    position of fourth point.

    In the following picture, `c1` correspond to "A", `c2` correspond to "B",
    `c3` correspond to "C", `d0` correspond to "L0", `return` correspond to "D".

    ![plpp](img/plpp.png)

    Set `inverse` option to `True` can make the result to the another side
    between `c1` and line `c2` `c3`.
    """
    cdef CCoord c = cplpp(CCoord(c1.x, c1.y), d0, CCoord(c2.x, c2.y),
                          CCoord(c3.x, c3.y), inverse)
    return Coord.__new__(Coord, c.x, c.y)


def palp(
    Coord c1,
    double a0,
    double d0,
    Coord c2,
    bint inverse = False
):
    """The PALP function requires two points, one angle and one distance,
    obtained the position of fourth point.

    In the following picture, `c1` correspond to "A", `c2` correspond to "B",
    `d0` correspond to "L0", `a0` correspond to "alpha", `return` correspond
    to "C".

    ![palp](img/palp.png)

    Set `inverse` option to `True` can make the result upside down.
    """
    cdef CCoord c = cpalp(CCoord(c1.x, c1.y), a0, d0, CCoord(c2.x, c2.y), inverse)
    return Coord.__new__(Coord, c.x, c.y)


cpdef int vpoint_dof(object vpoints):
    """Return the DOF of the mechanism expression `vpoints`."""
    # Joint with DOF 1
    cdef int j1 = 0
    # Joint with DOF 2
    cdef int j2 = 0
    # First link is frame
    vlinks = {VLink.FRAME}
    cdef int link_count
    cdef VPoint vp
    for vp in vpoints:
        link_count = len(vp.links)
        if link_count < 2:
            # If a point doesn't have two more links, it can not be call a 'joint'.
            continue
        vlinks.update(vp.links)
        if vp.type in {VJoint.R, VJoint.P}:
            j1 += link_count - 1
        elif vp.type == VJoint.RP:
            j1 += link_count - 2
            j2 += 1
    return 3 * (len(vlinks) - 1) - 2 * j1 - j2


cdef inline int base_friend(int node, object vpoints):
    cdef int i
    cdef VPoint vpoint
    for i, vpoint in enumerate(vpoints):
        if not vpoints[node].links:
            continue
        if vpoints[node].links[0] in vpoint.links:
            return i


cdef bint preprocessing(EStack exprs, object vpoints, object inputs,
                        map[Sym, CCoord] &joint_pos,
                        map[SwappablePair, double] &link_len,
                        map[Sym, double] &param):
    """Data extraction. Return true if input angle is matched DOF.

    Use "exprs", "vpoints", "angles" and "link_len" to generate solver
    required data. Please pre-allocate the "joint_pos" and "param".

    C++ objects
    + "exprs.stack" used for the position solution.
    + "joint_pos" used for joint position.
    + "link_len" used for custom link length except from calculation.
      (The link pairs must be exist in the solution.)
    + "param" used for store angles and link length.
    """
    vpoints = list(vpoints)
    cdef int vp_dof = vpoint_dof(vpoints)
    # First, we create a "VLinks" that can help us to
    # find a relationship just like adjacency matrix
    cdef int node, base
    cdef VPoint vp, vp2
    vlinks = {}
    for node, vp in enumerate(vpoints):
        for link in vp.links:
            # Add as vlink.
            if link not in vlinks:
                vlinks[link] = [node]
            else:
                vlinks[link].append(node)
    # Replace the P joints and their friends with RP joint
    # DOF must be same after properties changed
    cdef double x, y, angle
    for base in range(len(vpoints)):
        vp = vpoints[base]
        if vp.type != VJoint.P:
            continue
        for link in vp.links[1:]:
            for node in vlinks[link]:
                vp2 = vpoints[node]
                if node == base or vp2.is_slider():
                    continue
                x = vp2.c[0, 0]
                y = vp2.c[0, 1]
                vpoints[node] = VPoint.c_slider_joint([vp.links[0]] + [
                    link_ for link_ in vp2.links
                    if link_ not in vp.links
                ], VJoint.RP, vp.angle, x, y)
    # Assign joint positions
    for node, vp in enumerate(vpoints):
        base = 1 if vp.is_slider() else 0
        joint_pos[Sym(P_LABEL, node)] = CCoord(vp.c[base, 0], vp.c[base, 1])
    # Add slider slot virtual coordinates
    for node, vp in enumerate(vpoints):
        # PLPP dependencies
        if vp.type != VJoint.RP:
            continue
        base = base_friend(node, vpoints)
        angle = (vp.angle
                 - vp.slope_angle(vpoints[base], 1, 0)
                 + vp.slope_angle(vpoints[base], 0, 0)) / 180 * M_PI
        joint_pos[Sym(S_LABEL, node)] = CCoord(vp.c[1, 0] + cos(angle),
                                               vp.c[1, 1] + sin(angle))
    # Scan again to check if the parameter exists
    # Especially link lengths and angles
    cdef int dof = 0
    cdef Expr e
    cdef SwappablePair pair1, pair2
    for e in exprs.stack:
        pair1 = SwappablePair(e.c1.second, e.target.second)
        pair2 = SwappablePair(e.c2.second, e.target.second)
        if e.func == PXY:
            param[e.v1] = joint_pos[e.target].x - joint_pos[e.c1].x
            param[e.v2] = joint_pos[e.target].y - joint_pos[e.c1].y
            continue
        if e.func in {PLLP, PALP}:
            param[e.v2] = (link_len[pair2]
                           if link_len.find(pair2) != link_len.end() else
                           distance(joint_pos[e.c2].x,
                                    joint_pos[e.c2].y,
                                    joint_pos[e.target].x,
                                    joint_pos[e.target].y))
        if e.func in {PLLP, PLA, PLAP, PLPP}:
            param[e.v1] = (link_len[pair1]
                           if link_len.find(pair1) != link_len.end() else
                           distance(joint_pos[e.c1].x,
                                    joint_pos[e.c1].y,
                                    joint_pos[e.target].x,
                                    joint_pos[e.target].y))
            if e.func in {PLA, PLAP}:
                if e.v2.first == I_LABEL:
                    # Input angles
                    angle = inputs[pair1.first, pair1.second]
                    param[e.v2] = angle / 180 * M_PI
                    dof += 1
                else:
                    # A_LABEL
                    param[e.v2] = slope_angle(joint_pos[e.c1].x,
                                              joint_pos[e.c1].y,
                                              joint_pos[e.target].x,
                                              joint_pos[e.target].y)
                    if e.func == PLAP:
                        param[e.v2] -= slope_angle(joint_pos[e.c1].x,
                                                   joint_pos[e.c1].y,
                                                   joint_pos[e.c2].x,
                                                   joint_pos[e.c2].y)
                if (
                    (<VPoint>vpoints[e.c1.second]).grounded()
                    and (<VPoint>vpoints[e.target.second]).grounded()
                ):
                    raise ValueError("wrong driver definition")
    return dof <= len(inputs) <= vp_dof


cpdef list expr_solving(
    EStack exprs,
    object vpoints,
    object inputs = None
):
    """Solver function of Triangular method and BFGS method, for mechanism
    expression `vpoints`.

    The triangle expression stack `expr` is generated from
    [`t_config`](#t_config).

    Solver function will not handle slider input pairs in argument `angles`,
    which is only support revolute joints. In another way, the slider input
    pairs can be set by [`VPoint.disable_offset()`](#vpointdisable_offset)
    method.
    """
    # Blank sequences
    if inputs is None:
        inputs = {}
    cdef map[Sym, CCoord] joint_pos
    cdef map[SwappablePair, double] link_len
    cdef map[Sym, double] param
    if not preprocessing(exprs, vpoints, inputs, joint_pos, link_len, param):
        raise ValueError("wrong number of input parameters")
    # Check coverage
    status = {i: False for i in range(len(vpoints))}
    cdef Expr e
    for e in exprs.stack:
        status[e.c1.second] = status[e.target.second] = True
        if e.func in {PLAP, PLLP, PLPP, PALP}:
            status[e.c2.second] = True
            if e.func == PLPP:
                status[e.c3.second] = True
    cdef bint bfgs_mode = not all(status.values())
    # Solve
    cdef ExprSolver solver = ExprSolver(exprs.stack, joint_pos, param)
    if not solver.solve():
        raise ValueError("solve failed")
    # Use BFGS mode
    cdef pair[Sym, CCoord] jp
    if bfgs_mode:
        data_dict = {}
        for jp in solver.joint_pos:
            if (
                jp.first.first == P_LABEL
                and status[jp.first.second]
            ):
                data_dict[jp.first.second] = Coord.__new__(Coord, jp.second.x,
                                                           jp.second.y)
        return SolverSystem(vpoints, inputs, data_dict).solve()
    rt = []
    cdef int i
    cdef CCoord c
    cdef VPoint vp
    for i, vp in enumerate(vpoints):
        c = solver.joint_pos[Sym(P_LABEL, i)]
        if vp.is_slider():
            rt.append(((vp.c[0, 0], vp.c[0, 1]), (c.x, c.y)))
        else:
            rt.append((c.x, c.y))
    return rt


cdef (bint, map[Sym, CCoord]) quick_solve(
    vector[Expr] stack,
    map[Sym, CCoord] joint_pos,
    map[Sym, double] param
) nogil:
    """Quick solving function.

    !! Cython can not expose C external function to another pyx.
    """
    cdef ExprSolver s = ExprSolver(stack, joint_pos, param)
    cdef bint ok = s.solve()
    return ok, s.joint_pos


def uniform_four_bar(double ml, int n):
    """Generate n four bar mechanisms from maximum lengths.

    These mechanisms have coupling points.
    Normalized parameters are $[L_0, L_2, L_3, L_4, \\alpha]$.

    ![pxy](img/uniform_four_bar.png)
    """
    return array(_uniform_four_bar(ml, n))


cdef double[:, :] _uniform_four_bar(double ml, int n):
    """Uniform four-bar implementation."""
    cdef double[:, :] d = uniform(1., ml, (n, 5))
    cdef int i
    for i in range(n):
        while not d[i, 0] + d[i, 2] > 1 + d[i, 1]:
            for j in range(4):
                d[i, j] = uniform(1., ml)
        d[i, 4] = uniform(1., 2 * M_PI)
    return d


def uniform_path(double[:, :] v, int n):
    """Generate path with four-bar dimensions.

    Normalized parameters are $[L_0, L_2, L_3, L_4, \\alpha]$.
    """
    return array(c_uniform_path(v, n))


cdef (vector[Expr], map[Sym, CCoord], map[Sym, double]) ueb(double[:] v) nogil:
    """Uniform expression builder."""
    cdef vector[Expr] stack
    stack.push_back(Expr(False, PLA, Sym(L_LABEL, 1), Sym(I_LABEL, 0),
                         Sym(P_LABEL, 0), Sym(), Sym(), Sym(P_LABEL, 2)))
    stack.push_back(Expr(False, PLLP, Sym(L_LABEL, 2), Sym(L_LABEL, 3),
                         Sym(P_LABEL, 2), Sym(P_LABEL, 1), Sym(),
                         Sym(P_LABEL, 3)))
    stack.push_back(Expr(False, PLAP, Sym(L_LABEL, 4), Sym(A_LABEL, 0),
                         Sym(P_LABEL, 2), Sym(P_LABEL, 3), Sym(),
                         Sym(P_LABEL, 4)))
    cdef map[Sym, CCoord] joint_pos
    joint_pos[Sym(P_LABEL, 0)] = CCoord(0, 0)
    cdef map[Sym, double] param
    param[Sym(L_LABEL, 1)] = 1.
    joint_pos[Sym(P_LABEL, 1)] = CCoord(v[0], 0)
    cdef int i
    for i in range(1, 4):
        param[Sym(L_LABEL, i + 1)] = v[i]
    param[Sym(A_LABEL, 0)] = v[4]
    # Assign driver angle
    # param[Sym(I_LABEL, 0)] = a
    return stack, joint_pos, param


cdef double[:, :, :] c_uniform_path(double[:, :] v, int n) nogil:
    """Uniform path implementation."""
    cdef double[:, :, :] p
    with gil:
        p = zeros((v.shape[0], n, 2))
    cdef vector[Expr] stack
    cdef map[Sym, CCoord] joint_pos
    cdef map[Sym, double] param
    cdef bint ok
    cdef int i, j
    cdef double a
    cdef CCoord c
    cdef map[Sym, CCoord] ans
    for i in range(len(v)):
        stack, joint_pos, param = ueb(v[i])
        j = 0
        a = 0
        while a < 2 * M_PI:
            param[Sym(I_LABEL, 0)] = a
            ok, ans = quick_solve(stack, joint_pos, param)
            if ok:
                c = ans[Sym(P_LABEL, 4)]
                p[i, j, 0] = c.x
                p[i, j, 1] = c.y
            else:
                p[i, j, 0] = NAN
                p[i, j, 1] = NAN
            a += 2 * M_PI / n
            j += 1
    return p


cpdef object uniform_expr(double[:] v):
    """Turn the uniform link length into expression."""
    cdef vector[Expr] stack
    cdef map[Sym, CCoord] joint_pos
    cdef map[Sym, double] param
    stack, joint_pos, param = ueb(v)
    cdef vector[CCoord] coords = vector[CCoord](3, CCoord(0., 0.))
    cdef bint ok = False
    cdef double a = 0
    cdef map[Sym, CCoord] ans
    while a < 2 * M_PI:
        param[Sym(I_LABEL, 0)] = a
        ok, ans = quick_solve(stack, joint_pos, param)
        for i in range(2, 5):
            coords[i - 2] = ans[Sym(P_LABEL, i)]
        if ok:
            break
        a += 2 * M_PI / 30
    return [
        VPoint.c_r_joint([VLink.FRAME, 'L1'], 0., 0.),
        VPoint.c_r_joint([VLink.FRAME, 'L3'], v[0], 0.),
        VPoint.c_r_joint(['L1', 'L2'], coords[0].x, coords[0].y),
        VPoint.c_r_joint(['L2', 'L3'], coords[1].x, coords[1].y),
        VPoint.c_r_joint(['L2'], coords[2].x, coords[2].y),
    ]



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Wrapper header of BFGS algorithm.
If you need to use container to carry floating point data,
std::list is recommended instead of std::vector.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cdef extern from "bfgs_solver/solve.h" nogil:
    struct Point:
        double *x
        double *y

    struct Line:
        Point *p1
        Point *p2

    struct Constraint:
        unsigned type
        Point *point1
        Point *point2
        Line *line1
        Line *line2
        double *parameter

    Constraint point_on_point(Point *, Point *)
    Constraint p2p_distance(Point *, Point *, double *)
    Constraint point_on_line(Point *, Line *)
    Constraint internal_angle(Line *, Line *, double *)
    Constraint line_internal_angle(Line *, double *)
    bint solve(double **, size_t, Constraint *, size_t, bint)



# -*- coding: utf-8 -*-

"""Kernel of Metaheuristic Algorithm."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from typing import Mapping, Dict, Union, Type
from enum import unique, Enum
from .utility import ObjFunc, Algorithm
from .config_types import (
    Setting, DESetting, FASetting, RGASetting, TOBLSetting,
)
from .rga import RGA
from .de import DE
from .pso import PSO
from .fa import FA
from .tlbo import TLBO


@unique
class AlgorithmType(str, Enum):
    """Enum type of algorithms."""
    RGA = "Real-coded Genetic Algorithm"
    DE = "Differential Evolution"
    PSO = "Particle Swarm Optimization"
    FA = "Firefly Algorithm"
    TLBO = "Teaching Learning Based Optimization"


_ALGORITHM: Mapping[AlgorithmType, Type[Algorithm]] = {
    AlgorithmType.RGA: RGA,
    AlgorithmType.DE: DE,
    AlgorithmType.PSO: PSO,
    AlgorithmType.FA: FA,
    AlgorithmType.TLBO: TLBO,
}
_DEFAULT_PARAMS = {'max_gen': 1000, 'report': 50}
_PARAMS: Mapping[AlgorithmType, Dict[str, Union[int, float]]] = {
    AlgorithmType.RGA: {
        'pop_num': 500,
        'cross': 0.95,
        'mutate': 0.05,
        'win': 0.95,
        'delta': 5.,
    },
    AlgorithmType.DE: {
        'pop_num': 400,
        'strategy': 1,
        'f': 0.6,
        'cr': 0.9,
    },
    AlgorithmType.PSO: {
        'pop_num': 200,
        'cognition': 2.05,
        'social': 2.05,
        'velocity': 1.3,
    },
    AlgorithmType.FA: {
        'pop_num': 80,
        'alpha': 0.01,
        'beta_min': 0.2,
        'gamma': 1.,
        'beta0': 1.,
    },
    AlgorithmType.TLBO: {
        'pop_num': 50,
    },
}


def algorithm(opt: AlgorithmType) -> Type[Algorithm]:
    """Return the class of the algorithms."""
    return _ALGORITHM[opt]


def default(opt: AlgorithmType) -> Dict[str, Union[int, float]]:
    """Return the default settings of the algorithms."""
    config = _PARAMS[opt].copy()
    config.update(_DEFAULT_PARAMS.copy())
    return config



# -*- coding: utf-8 -*-

from typing import Tuple, List, Sequence, Callable, Optional
from .graph import Graph


def link_synthesis(
    nl: int,
    nj: int,
    stop_func: Optional[Callable[[], bool]] = None
) -> List[Tuple[int, ...]]:
    ...


def contracted_link_synthesis(
    link_num_list: Sequence[int],
    stop_func: Optional[Callable[[], bool]] = None
) -> List[Tuple[int, ...]]:
    ...

def conventional_graph(
    cg_list: List[Graph],
    c_j_list: Sequence[int],
    no_degenerate: int = 1,
    stop_func: Optional[Callable[[], bool]] = None
) -> List[Graph]:
    ...

def contracted_graph(
    link_num: Sequence[int],
    stop_func: Optional[Callable[[], bool]] = None
) -> List[Graph]:
    ...



# -*- coding: utf-8 -*-

from typing import Tuple, List, Iterable, Dict
from numpy import ndarray

def link_assortment(g: Graph) -> List[int]:
    ...

def contracted_link_assortment(g: Graph) -> List[int]:
    ...

def labeled_enumerate(g: Graph) -> List[Tuple[int, Graph]]:
    ...

class Graph:

    edges: Tuple[Tuple[int, int], ...]
    vertices: Tuple[int, ...]

    def __init__(self, edges: Iterable[Tuple[int, int]]):
        """Input edges of the graph. The vertices symbols are
        positive continuously integer.
        """
        ...

    def add_edge(self, n1: int, n2: int) -> None:
        ...

    def add_vertices(self, vertices: Iterable[int]) -> None:
        ...

    def dof(self) -> int:
        ...

    def neighbors(self, n: int) -> Tuple[int, ...]:
        ...

    def degrees(self) -> Dict[int, int]:
        ...

    def degree_code(self) -> int:
        ...

    def adjacency_matrix(self) -> ndarray:
        ...

    def is_connected(self, without: int = -1) -> bool:
        ...

    def has_cut_link(self) -> bool:
        ...

    def is_degenerate(self) -> bool:
        ...

    def has_triangle(self) -> bool:
        ...

    def is_isomorphic(self, graph: Graph) -> bool:
        ...

    def is_isomorphic_vf2(self, graph: Graph) -> bool:
        ...

    def is_isomorphic_degree_code(self, graph: Graph) -> bool:
        ...

    def duplicate(self, vertices: Iterable[int], times: int) -> Graph:
        ...

    def copy(self) -> Graph:
        ...



# -*- coding: utf-8 -*-

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from pyslvs.metaheuristics import algorithm, default, AlgorithmType
from pyslvs.metaheuristics.test import TestObj
from . import TestBase


class AlgorithmTest(TestBase):

    def case(self, alg: AlgorithmType):
        """Test with an objective function."""
        settings = {'min_fit': 1e-20, 'report': 10}
        obj = TestObj()
        s = default(alg)
        s.update(settings)
        a = algorithm(alg)(obj, s)
        ans = a.run()
        x, y = a.result()
        self.assertTrue(ans < 1e-20)
        for i in range(4):
            self.assertTrue(x[i] < 1e-10)
        self.assertEqual(y, ans)

    def test_rga(self):
        self.case(AlgorithmType.RGA)

    def test_de(self):
        self.case(AlgorithmType.DE)

    def test_pso(self):
        self.case(AlgorithmType.PSO)

    def test_fa(self):
        self.case(AlgorithmType.FA)

    def test_tlbo(self):
        self.case(AlgorithmType.TLBO)



# -*- coding: utf-8 -*-
# cython: language_level=3

"""PMKS symbolics.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cdef double distance(double x1, double y1, double x2, double y2) nogil
cdef double slope_angle(double x1, double y1, double x2, double y2) nogil
cpdef list get_vlinks(object vpoints)

cdef struct CCoord:
    double x, y

cdef class Coord:
    # Coordinate data type.
    # Needs to replaced at backend with c-struct type CCoord.
    cdef public double x, y

    cpdef double distance(self, Coord p)
    cpdef double slope_angle(self, Coord p)
    cpdef bint is_nan(self)

cpdef enum VJoint:
    # Joint types.
    # Actually "class VJoint(IntEnum)" in Python but "enum" in C++.
    R  # Rotate pair
    P  # Prismatic pair
    RP  # Rotate and prismatic pair

cdef class VPoint:
    # VPoint(links, type_int, angle, color_str, x, y, color_func=None)

    cdef readonly tuple links
    cdef readonly double[:, :] c
    cdef readonly VJoint type
    cdef readonly tuple color
    cdef readonly str color_str
    cdef readonly str type_str
    cdef readonly double x, y, angle
    cdef double __offset
    cdef bint __has_offset

    @staticmethod
    cdef VPoint c_r_joint(object links, double x, double y)
    @staticmethod
    cdef VPoint c_slider_joint(object links, VJoint type_int, double angle, double x, double y)

    # Copy method
    cpdef VPoint copy(self)

    # Set values
    cpdef void set_links(self, object links) except *
    cpdef void replace_link(self, str link1, str link2) except *
    cpdef void move(self, object c1, object c2 = *) except *
    cpdef void locate(self, double x, double y) except *
    cpdef void rotate(self, double angle)
    cpdef void set_offset(self, double offset)
    cpdef void disable_offset(self)

    # Get or calculate values
    cpdef bint is_slider(self)
    cpdef double distance(self, VPoint p)
    cpdef bint has_offset(self)
    cpdef double offset(self)
    cpdef double true_offset(self)
    cpdef double slope_angle(self, VPoint p, int num1 = *, int num2 = *)
    cpdef Coord link_pos(self, str link)

    # Link operators
    cpdef bint grounded(self)
    cpdef bint pin_grounded(self)
    cpdef bint same_link(self, VPoint p)
    cpdef bint no_link(self)
    cpdef bint is_slot_link(self, str link)

    # Expression
    cpdef str expr(self)
    cpdef Coord to_coord(self, size_t ind)

cdef class VLink:
    # VLink(name, color_str, points, color_func=None)

    cdef readonly str name, color_str
    cdef readonly tuple color
    cdef readonly long[:] points

    # Set values
    cpdef void set_points(self, object points) except *
    cpdef Coord[:] points_pos(self, object vpoints) except *



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Structural synthesis.

The algorithm reference:
+ On the Number Synthesis of Kinematic Chains
    + author: Hong-Sen Yan, Yu-Ting Chiu
    + Mechanism and Machine Theory, Volume 89, 2015, Pages 128-144, ISSN 0094-114X
    + https://doi.org/10.1016/j.mechmachtheory.2014.08.012

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from time import process_time
from logging import getLogger
from collections import Counter
cimport cython
from libcpp.pair cimport pair
from libcpp.vector cimport vector
from libcpp.map cimport map
from numpy import (
    int16 as i16,
    array as np_array,
    zeros as np_zeros,
    ones as np_ones,
    sum as np_sum,
    subtract as np_sub,
    multiply as np_mul,
    floor_divide as np_div,
    any as np_any,
    prod as np_prod,
    repeat as np_repeat,
    arange,
)
from .graph cimport Graph, link_assortment
from .planar cimport is_planar

ctypedef unsigned int uint
ctypedef unsigned long long ullong
ctypedef pair[int, int] ipair
ctypedef map[int, int] imap

logger = getLogger()


@cython.boundscheck(False)
@cython.wraparound(False)
cdef short[:, :] _product(tuple pool, object stop_func):
    """Product function as same as iteration tools.

    The pool is created by range(n).
    """
    if not pool:
        return np_array([], dtype=i16)
    cdef short[:] tmp1
    cdef short[:, :] tmp2
    cdef short[:] array0 = arange(pool[0], dtype=i16)
    cdef int n = np_prod(pool)
    cdef short[:, :] out = np_zeros((n, len(pool)), dtype=i16)
    cdef int array0_size = len(array0)
    cdef int m = n / array0_size
    tmp1 = np_repeat(array0, m)
    out[:, 0] = tmp1
    cdef int j
    if pool[1:]:
        tmp2 = _product(pool[1:], stop_func)
        out[0:m, 1:] = tmp2
        for j in range(1, array0_size):
            if stop_func is not None and stop_func():
                return out
            out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out


cdef short[:, :] _simple_product(int pool_size, int repeat, object stop_func):
    return _product((pool_size,) * repeat + (1,), stop_func)


cdef inline int _m_max(int nl, int nj) nogil:
    """Find max number of joint on each link.

    + nl <= nj and nj <= (2 * nl - 3)
    + (2 * nl - 3) <= nj and nj <= (nl * (nl - 1) / 2)
    + other exceptions (return -1).
    """
    if nl <= nj <= (2 * nl - 3):
        return nj - nl + 2
    if nl == nj == 0:
        return -1
    if 2 * nl - 3 <= nj <= <double>(nl * (nl - 1)) / 2:
        return nl - 1
    return -1


cdef inline int _sum_factors(short[:] factors):
    """F0*N2 + F1*N3 + F2*N4 + ... + Fn*N(n+2)"""
    cdef int factor = 0
    cdef int i
    for i in range(len(factors)):
        factor += factors[i] * (i + 2)
    return factor


cpdef list link_synthesis(int nl, int nj, object stop_func = None):
    """Return link assortment by number of links `nl` and number of joints `nj`.

    The check stop function `stop_func` object for GUI or subprocess,
    return `True` to terminate this function.
    """
    result = []
    cdef int m_max_v = _m_max(nl, nj)
    if m_max_v == -1:
        raise ValueError("incorrect mechanism.")

    cdef int i, p
    cdef short[:] symbols
    for symbols in _simple_product(nl + 1, m_max_v - 2, stop_func):
        nl_m_max = nl - np_sum(symbols)
        if nl_m_max < 0:
            continue
        symbols[-1] = nl_m_max
        if _sum_factors(symbols) == nj * 2:
            result.append(tuple(symbols))
    return result


cdef inline int _j_m(short[:] link_num):
    """Return value of JM."""
    cdef int num
    cdef int i = 3
    cdef float c = 0
    for num in link_num[1:]:
        c += <double>i / 2 * num
        i += 1
    return <int>c


cdef inline int _j_m_p(int n_m) nogil:
    """Return value of J'M. This is improved function.

    + Origin equation:
    if n_m % 2 == 0:
        return <int>((3 * (n_m - 1) - 1) / 2)
    else:
        return <int>((3 * (n_m - 1) - 2) / 2)
    """
    # Number of multiple links.
    if n_m <= 1:
        return 0
    elif n_m == 2:
        return 1
    else:
        return 3 * (n_m - 2)


cpdef list contracted_link_synthesis(object link_num_list, object stop_func = None):
    """Return contracted link assortment by link assortment `link_num_list`.

    The check stop function `stop_func` object for GUI or subprocess,
    return `True` to terminate this function.
    """
    cdef short[:] link_num
    if len(link_num_list) == 1:
        link_num = np_zeros(link_num_list[0], dtype=i16)
        link_num[-1] = 1
        return [tuple(link_num)]

    link_num = np_array(link_num_list, dtype=i16)

    # Contracted link
    cdef int j_m_v = _j_m(link_num)
    cdef int n_c_min = max(1, j_m_v - _j_m_p(sum(link_num[1:])))
    cdef int n_c_max = min(link_num[0], j_m_v)

    # i = NL2 - NC + 2
    cdef int i_max = min(link_num[0], link_num[0] - n_c_min + 2)

    # Matching formula
    cdef int count, factor, index
    cdef float last_factor
    cdef short[:] m
    cj_list = []
    for m in _simple_product(link_num[0] + 1, i_max - 1, stop_func):
        count = 0
        index = 1
        for factor in m[:-1]:
            count +=  factor * index
            index += 1

        # Check if the last factor is a natural number.
        last_factor = <double>(link_num[0] - count) / index
        factor = <int>last_factor
        if last_factor < 0 or last_factor != factor:
            continue

        m[-1] = factor
        if n_c_min <= np_sum(m) <= n_c_max:
            cj_list.append(tuple(m))

    return cj_list


cdef inline short[:] _nonzero_index(short[:] array):
    """Return the number of nonzero numbers."""
    counter = []
    cdef int i, n
    for i, n in enumerate(array):
        if n != 0:
            counter.append(i)
    return np_array(counter, dtype=i16)


cdef inline ullong _factorial(int n):
    """Return (n!)."""
    cdef ullong ret = 1
    cdef int i
    for i in range(2, n + 1):
        ret *= i
    return ret


cdef inline int _gcd(int a, int b) nogil:
    """Return Greatest Common Divisor of a and b.

    Only for positive numbers.
    """
    cdef int r
    while b > 0:
        r = a % b
        a = b
        b = r
    return a


cdef inline short[:] _labels(short[:] numbers, int index, int offset):
    """Generate labels from numbers."""
    cdef int i, num
    labels = []
    for num in numbers[offset:]:
        for i in range(num):
            labels.append(index)
        index += 1
    return np_array(labels, dtype=i16)


cdef inline Graph _multigraph(short[:] counter, int n):
    """Get multigraph from n x n matrix."""
    edges = {}
    cdef int c = 0
    cdef int i, j
    for i in range(n):
        for j in range(n):
            if i >= j:
                continue
            edges[i, j] = counter[c]
            c += 1
    return Graph(Counter(edges).elements())


@cython.boundscheck(False)
cdef inline void _gauss_elimination(
    list result,
    short[:] limit,
    short[:, :] f_matrix,
    int n,
    int var_count
):
    """Gauss elimination for (n, n + 1) matrix."""
    cdef int i, j, d
    cdef short[:] tmp1, tmp2
    for j in range(var_count):
        # Remove all coefficients of index [i] to zero
        for i in range(n):
            if f_matrix[i, j] != 0 and not np_any(f_matrix[i, :j]):
                d = i
                tmp2 = np_div(f_matrix[i, :], f_matrix[i, j])
                break
        else:
            continue

        for i in range(n):
            if i == d or f_matrix[i, j] == 0:
                continue
            tmp1 = np_sub(f_matrix[i, :], np_mul(tmp2, f_matrix[i, j]))
            f_matrix[i, :] = tmp1

    # Answer
    cdef short[:] answer = -np_ones(var_count, dtype=i16)

    # Determined solution
    cdef int c, k
    for i in range(n):
        c = 0
        for j in range(var_count):
            # Derivation (has answer)
            if answer[j] != -1 and f_matrix[i, j] != 0:
                f_matrix[i, -1] -= f_matrix[i, j] * answer[j]
                f_matrix[i, j] = 0

            # Nonzero coefficient
            if f_matrix[i, j] != 0:
                d = j
                c += 1

        if c != 1:
            continue

        j = d
        k = f_matrix[i, j]
        c = f_matrix[i, -1]
        if k != 1:
            if k < 0:
                k = -k
                c = -c
            if c < 0:
                return
            d = _gcd(k, c)
            k /= d
            c /= d
        if c < 0:
            return
        answer[j] = c

    # Result
    _test_contracted_graph(_multigraph(answer, n), result)


@cython.boundscheck(False)
cdef void _nest_do(
    list result,
    short[:] answer,
    short[:, :] f_matrix,
    int i,
    int n,
    object stop_func
):
    """Nest do loop."""
    if i >= n:
        # Result
        if <int>np_sum(answer) == <int>np_sum(f_matrix[:, -1]) / 2:
            _test_contracted_graph(_multigraph(answer, n), result)
        return

    cdef short[:] coefficients = _nonzero_index(f_matrix[i, :-1])
    cdef int c = 0
    cdef int j = -1
    cdef int d
    for d in coefficients:
        if answer[d] == -1:
            c += 1
            j = d

    if c == 0:
        if <int>np_sum(np_mul(answer, f_matrix[i, :-1])) == f_matrix[i, -1]:
            _nest_do(result, answer, f_matrix, i + 1, n, stop_func)
        return
    elif c == 1:
        c = 0
        for d in coefficients:
            if answer[d] != -1:
                c += answer[d] * f_matrix[i, d]
        answer[j] = (f_matrix[i, -1] - c) / f_matrix[i, j]
        _nest_do(result, answer, f_matrix, i + 1, n, stop_func)
        return

    cdef short[:] combine, answer_copy
    for combine in _product((f_matrix[i, -1],) * c, stop_func):
        if stop_func is not None and stop_func():
            return

        c = 0
        d = 0
        for j in coefficients:
            if answer[j] == -1:
                c += combine[d] * f_matrix[i, j]
                d += 1
            else:
                c += answer[j] * f_matrix[i, j]
        if c != f_matrix[i, -1]:
            continue

        answer_copy = answer.copy()
        d = 0
        for j in coefficients:
            # Pass to answer
            if answer_copy[j] == -1:
                answer_copy[j] = combine[d]
                d += 1
        _nest_do(result, answer_copy, f_matrix, i + 1, n, stop_func)


cdef inline bint _is_isomorphic(Graph g, list result):
    """Return true if graph is isomorphic with result list."""
    cdef Graph h
    for h in result:
        if g.is_isomorphic(h):
            return True
    return False


cdef inline void _test_contracted_graph(Graph g, list result):
    """Test the contracted graph."""
    if not g.edges:
        return
    # All connected
    if not g.is_connected():
        return
    # Cut link
    if g.has_cut_link():
        return
    # Planar graph
    if not is_planar(g):
        return
    # Isomorphism
    if _is_isomorphic(g, result):
        return

    result.append(g)


cdef inline void _test_graph(
    Graph g,
    list result,
    uint no_degenerate
):
    """Test result graph."""
    # Graph filter depending on degenerate option
    if no_degenerate == 0 and not g.is_degenerate():
        return
    elif no_degenerate == 1 and g.is_degenerate():
        return
    # Isomorphism
    if _is_isomorphic(g, result):
        return
    result.append(g)


@cython.boundscheck(False)
cdef inline void _contracted_graph(
    list result,
    short[:] limit,
    object stop_func
):
    """Synthesis of contracted graphs."""
    cdef int n = len(limit)
    cdef int var_count = n * (n - 1) / 2
    cdef short[:, :] f_matrix = np_zeros((n, var_count + 1), dtype=i16)
    f_matrix[:, -1] = limit

    # Equations
    cdef int i, j, k, c
    for i in range(n):
        c = 0
        for j in range(n):
            for k in range(n):
                if j >= k:
                    continue
                if i in {j, k}:
                    f_matrix[i, c] = 1
                c += 1

    if n >= var_count:
        # Fast solution by Gauss elimination.
        _gauss_elimination(result, limit, f_matrix, n, var_count)
    else:
        # Nest do loop method.
        _nest_do(result, -np_ones(var_count, dtype=i16), f_matrix, 0, n, stop_func)


cdef inline void _dyad_insert(Graph g, frozenset edge, int amount):
    """Insert dyad to the graph."""
    if amount < 1:
        return

    cdef int last_num = max(g.vertices) + 1
    cdef int n1, n2
    n1, n2 = edge

    path = [n1]
    path.extend(range(last_num, last_num + amount))
    path.append(n2)

    g.remove_edge(n1, n2)
    g.add_path(path)


cdef inline void _permute_combine(
    short[:] limit,
    list combine_list,
    tuple pick_list,
    object stop_func
):
    """Permutation of combined list."""
    cdef int n = len(limit)
    if n < 1:
        return

    cdef vector[int] indices = range(n)
    cdef short[:] cycles = np_zeros(n, dtype=i16)
    cdef short[:] pool = limit
    permute_list = set()
    cdef int i, j
    for i, j in enumerate(range(n, 0, -1)):
        cycles[i] = j
    permute_list.add(tuple(pool[indices[i]] for i in range(n)))
    cdef vector[int].iterator it2 = indices.begin()

    cdef int tmp
    while True:
        if stop_func is not None and stop_func():
            break

        for i in reversed(range(n)):
            cycles[i] -= 1
            if cycles[i] == 0:
                tmp = indices[i]
                indices.erase(it2 + i)
                indices.push_back(tmp)
                cycles[i] = n - i
            else:
                j = cycles[i]
                tmp = indices[i]
                indices[i] = indices[n - j]
                indices[n - j] = tmp
                permute_list.add(tuple(pool[indices[i]] for i in range(n)))
                break
        else:
            break
    for tmp_array in permute_list:
        combine_list.append(tuple(zip(pick_list, tmp_array)))


cdef inline list _contracted_links(tuple edges, short[:] limit, object stop_func):
    """Combination of contracted links.

    pool: edges
    pick: contracted links

    If the edge is not picked, it represent the joint is connected directly.
    """
    # Number of contracted links
    cdef int pick_count = len(limit)
    if pick_count < 1:
        return []

    # Check over picked
    pool_list = tuple([frozenset(edge) for edge in edges])
    pool = Counter(pool_list)
    # The list including required edge(s).
    confirm_list = pool - Counter(set(pool_list))
    # Simplified the pool
    pool_list = tuple((pool - confirm_list).elements())
    cdef int confirm_size = sum(confirm_list.values())
    if confirm_size > pick_count or pick_count > len(edges):
        return []

    pick_count -= confirm_size
    cdef int pool_size = len(pool_list)
    cdef short[:] indices = np_zeros(pick_count, dtype=i16)
    cdef int i
    for i in range(pick_count):
        indices[i] = i
    pick_list = Counter()
    combine_set = set()
    # Combinations loop with number checking.
    cdef int n1, n2
    while True:
        if stop_func is not None and stop_func():
            break

        # Combine
        for i in range(pick_count):
            pick_list[pool_list[indices[i]]] += 1

        # Collecting
        combine_set.add(tuple(pick_list.elements()))
        # Initialize
        pick_list.clear()
        # Check combination is over.
        for n1 in reversed(range(pick_count)):
            if indices[n1] != n1 + pool_size - pick_count:
                break
        else:
            break
        # Next indicator
        indices[n1] += 1
        for n2 in range(n1 + 1, pick_count):
            indices[n2] = indices[n2 - 1] + 1
    combine_list = []
    pool_list = tuple(confirm_list.elements())
    for picked in combine_set:
        _permute_combine(limit, combine_list, pool_list + picked, stop_func)
    return combine_list


cdef inline void _graph_atlas(
    list result,
    list contracted_graph,
    short[:] limit,
    uint no_degenerate,
    object stop_func
):
    """Synthesis of atlas."""
    cdef int n
    cdef Graph cg, g
    for cg in contracted_graph:
        for combine in _contracted_links(cg.edges, limit, stop_func):
            g = Graph.__new__(Graph, cg.edges)
            for edge, n in combine:
                _dyad_insert(g, edge, n)
            _test_graph(g, result, no_degenerate)


cdef inline list _loop_chain(int num):
    """Loop chain of num."""
    cdef int i
    cdef int b = 0
    chain = []
    for i in range(1, num):
        chain.append((b, i))
        b = i
    chain.append((0, b))
    return chain


cpdef list contracted_graph(object link_num_list, object stop_func = None):
    """Generate contracted graphs by link assortment `link_num`.

    The check stop function `stop_func` object for GUI or subprocess,
    return `True` to terminate this function.
    """
    if not link_num_list:
        return []

    # Initial time
    cdef double t0 = process_time()
    cdef short[:] link_num = np_array(link_num_list, ndmin=1, dtype=i16)
    logger.debug(f"Link assortment: {list(link_num)}")

    # Multiple links
    cdef short[:] m_link = _labels(link_num, 3, 1)

    cdef imap m_limit, count
    cdef int num
    cdef int i = 0
    for num in m_link:
        m_limit[i] = num
        count[i] = 0
        i += 1

    # Synthesis of contracted graphs
    cg_list = []
    _contracted_graph(cg_list, m_link, stop_func)
    logger.debug(f"Contracted graph(s): {len(cg_list)}, time: {process_time() - t0}")
    return cg_list


cpdef list conventional_graph(
    list cg_list,
    object c_j_list,
    uint no_degenerate = 1,
    object stop_func = None
):
    """Generate conventional graphs by contracted graphs `cg_list` and
    contracted link assortment `c_j_list`.

    The degenerate setting `no_degenerate` has following option:

    + `0`: No degenerate.
    + `1`: Only degenerate.
    + Else: All graphs.

    The check stop function `stop_func` object for GUI or subprocess,
    return `True` to terminate this function.
    """
    # Initial time
    cdef double t0 = process_time()
    logger.debug(f"Contracted link assortment: {list(c_j_list)}")

    # Synthesis of contracted link and multiple link combination.
    cdef short[:] c_j = np_array(c_j_list, ndmin=1, dtype=i16)
    result = []
    cdef int i, num
    if not cg_list:
        if 1 not in c_j_list:
            return []
        # Single loop - ring graph (special case)
        i = 1
        for num in c_j:
            if num == 1:
                break
            i += 1
        else:
            raise ValueError("Invalid assortment!")
        result.append(Graph.__new__(Graph, _loop_chain(i)))
        logger.debug(f"Count: {len(result)}")
        return result

    # Multiple links
    cdef short[:] m_link = np_array(link_assortment(cg_list[0]), ndmin=1, dtype=i16)
    m_link = _labels(m_link, 3, 1)

    # Synthesis of multiple links
    _graph_atlas(result, cg_list, _labels(c_j, 1, 0), no_degenerate, stop_func)

    # Return graph list and time
    logger.debug(f"Count: {len(result)}, time: {process_time() - t0}")
    return result



# -*- coding: utf-8 -*-
# cython: language_level=3

"""The callable class of the validation in algorithm.
The 'utility' module should be loaded when using sub-class.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from libc.time cimport time_t
from libcpp.list cimport list as clist

ctypedef unsigned int uint

cdef enum Task:
    MAX_GEN
    MIN_FIT
    MAX_TIME
    SLOW_DOWN

cdef packed struct Report:
    uint gen
    double fitness
    double time

cdef double rand_v(double lower = *, double upper = *) nogil
cdef uint rand_i(uint upper) nogil


cdef class ObjFunc:
    cdef uint gen
    cdef double[:] ub
    cdef double[:] lb

    cdef double fitness(self, double[:] v) nogil
    cpdef object result(self, double[:] v)


cdef class Algorithm:
    cdef bint parallel
    cdef uint pop_num, dim, rpt
    cdef Task task
    cdef double stop_at, best_f
    cdef double[:] best, fitness
    cdef double[:, :] pool
    cdef time_t time_start
    cdef clist[Report] reports
    cdef object progress_fun, interrupt_fun
    cdef public ObjFunc func

    # Chromosome
    cdef void assign(self, uint i, uint j) nogil
    cdef void assign_from(self, uint i, double f, double[:] v) nogil
    cdef void set_best(self, uint i) nogil
    cdef void find_best(self) nogil

    cdef void init_pop(self) nogil
    cdef void init(self) nogil
    cdef void generation(self) nogil
    cdef double check(self, int s, double v) nogil
    cdef void report(self) nogil
    cpdef double[:, :] history(self)
    cpdef tuple result(self)
    cpdef object run(self)



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""The callable class of the validation in algorithm.
The 'utility' module should be loaded when using sub-class of base classes.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from numpy import array, zeros, float64 as f64
from libc.math cimport HUGE_VAL
from libc.stdlib cimport rand, srand, RAND_MAX
from libc.time cimport time, difftime


cdef inline double rand_v(double lower = 0., double upper = 1.) nogil:
    """Random real value between lower <= r <= upper."""
    return lower + <double>rand() / RAND_MAX * (upper - lower)


cdef inline uint rand_i(uint upper) nogil:
    """A random integer between 0 <= r < upper."""
    return rand() % upper


cdef class ObjFunc:
    """Objective function base class.

    It is used to build the objective function for Meta-heuristic Algorithms.
    """

    cdef double fitness(self, double[:] v) nogil:
        with gil:
            raise NotImplementedError

    cpdef object result(self, double[:] v):
        """The result function. Default is the best variable vector `v`."""
        return array(v)


cdef class Algorithm:
    """Algorithm base class.

    It is used to build the Meta-heuristic Algorithms.
    """

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        """Generic settings."""
        srand(time(NULL))
        # object function
        self.func = func
        self.stop_at = 0
        if 'max_gen' in settings:
            self.task = MAX_GEN
            self.stop_at = settings['max_gen']
        elif 'min_fit' in settings:
            self.task = MIN_FIT
            self.stop_at = settings['min_fit']
        elif 'max_time' in settings:
            self.task = MAX_TIME
            self.stop_at = settings['max_time']
        elif 'slow_down' in settings:
            self.task = SLOW_DOWN
            self.stop_at = 1 - settings['slow_down']
        else:
            raise ValueError("please give 'max_gen', 'min_fit' or 'max_time' limit")
        self.pop_num = settings['pop_num']
        self.rpt = settings['report']
        if self.rpt <= 0:
            self.rpt = 10
        self.progress_fun = progress_fun
        self.interrupt_fun = interrupt_fun
        self.dim = len(self.func.ub)
        if self.dim != len(self.func.lb):
            raise ValueError("length of upper and lower bounds must be equal")
        self.fitness = zeros(self.pop_num, dtype=f64)
        self.pool = zeros((self.pop_num, self.dim), dtype=f64)
        self.best_f = HUGE_VAL
        self.best = zeros(self.dim, dtype=f64)
        # setup benchmark
        self.func.gen = 0
        self.time_start = 0
        self.reports = clist[Report]()

    cdef void assign(self, uint i, uint j) nogil:
        """Copy value from j to i."""
        self.fitness[i] = self.fitness[j]
        self.pool[i, :] = self.pool[j, :]

    cdef void assign_from(self, uint i, double f, double[:] v) nogil:
        """Copy value from tmp."""
        self.fitness[i] = f
        self.pool[i, :] = v

    cdef void set_best(self, uint i) nogil:
        """Set as best."""
        self.best_f = self.fitness[i]
        self.best[:] = self.pool[i, :]

    cdef void find_best(self) nogil:
        """Find the best."""
        cdef uint best = 0
        cdef uint i
        for i in range(0, self.pop_num):
            if self.fitness[i] < self.fitness[best]:
                best = i
        if self.fitness[best] < self.best_f:
            self.set_best(best)

    cdef void init_pop(self) nogil:
        """Initialize population."""
        cdef uint best = 0
        cdef uint i, s
        for i in range(self.pop_num):
            for s in range(self.dim):
                self.pool[i, s] = rand_v(self.func.lb[s], self.func.ub[s])
            self.fitness[i] = self.func.fitness(self.pool[i, :])
            if self.fitness[i] < self.fitness[best]:
                best = i
        if self.fitness[best] < self.best_f:
            self.set_best(best)

    cdef void init(self) nogil:
        """Initialize function."""
        pass

    cdef void generation(self) nogil:
        """The process of each generation."""
        with gil:
            raise NotImplementedError

    cdef double check(self, int s, double v) nogil:
        """Check the bounds."""
        if v > self.func.ub[s]:
            return self.func.ub[s]
        elif v < self.func.lb[s]:
            return self.func.lb[s]
        else:
            return v

    cdef inline void report(self) nogil:
        """Report generation, fitness and time."""
        self.reports.push_back(Report(
            self.func.gen,
            self.best_f,
            difftime(time(NULL), self.time_start),
        ))

    cpdef double[:, :] history(self):
        """Return the history of the process.

        The first value is generation (iteration);
        the second value is fitness;
        the third value is time in second.
        """
        return array([
            (report.gen, report.fitness, report.time)
            for report in self.reports
        ], dtype=f64)

    cpdef tuple result(self):
        """Return the best variable vector and its fitness."""
        return array(self.best), self.best_f

    cpdef object run(self):
        """Run and return the result and convergence history.

        The first place of `return` is came from
        calling [`ObjFunc.result()`](#objfuncresult).

        The second place of `return` is a list of generation data,
        which type is `Tuple[int, float, float]]`.
        The first of them is generation,
        the second is fitness, and the last one is time in second.
        """
        # Swap upper and lower bound if reversed
        for i in range(len(self.func.ub)):
            if self.func.ub[i] < self.func.lb[i]:
                self.func.ub[i], self.func.lb[i] = self.func.lb[i], self.func.ub[i]
        # Start
        self.time_start = time(NULL)
        self.init_pop()
        self.init()
        self.report()
        # Iterations
        cdef double diff, best_f
        cdef double last_diff = 0
        while True:
            best_f = self.best_f
            self.func.gen += 1
            self.generation()
            if self.func.gen % self.rpt == 0:
                self.report()
            if self.task == MAX_GEN:
                if self.func.gen >= self.stop_at > 0:
                    break
            elif self.task == MIN_FIT:
                if self.best_f <= self.stop_at:
                    break
            elif self.task == MAX_TIME:
                if difftime(time(NULL), self.time_start) >= self.stop_at > 0:
                    break
            elif self.task == SLOW_DOWN:
                diff = best_f - self.best_f
                if last_diff > 0 and diff / last_diff >= self.stop_at:
                    break
                last_diff = diff
            # progress
            if self.progress_fun is not None:
                self.progress_fun(self.func.gen, f"{self.best_f:.04f}")
            # interrupt
            if self.interrupt_fun is not None and self.interrupt_fun():
                break
        self.report()
        return self.func.result(self.best)



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Normalized planar four-bar linkage synthesis.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libc.math cimport (
    M_PI, INFINITY as INF, HUGE_VAL, isnan, hypot, atan2, cos, sin, sqrt,
)
from numpy import array, float64 as f64, hstack
from numpy.linalg import eig
from numpy.fft import fft
from pyslvs.expression cimport VPoint
from pyslvs.tinycadlib cimport c_uniform_path, uniform_expr
from pyslvs.metaheuristics.utility cimport ObjFunc
from .f_planar cimport roll


cdef (double, double) axes_v(double[:] v1, double[:] v2, double[:, :] p1,
                             double x_mean, double y_mean) nogil:
    """Calculate the orientation vector of the axes."""
    cdef double a = v1[1] / v1[0]
    cdef double b = y_mean - a * x_mean
    cdef double neg_x = 0
    cdef double neg_y = 0
    cdef double pos_x = 0
    cdef double pos_y = 0
    cdef int i
    cdef double val, x, y
    for i in range(len(p1)):
        val = p1[i, 1] - a * p1[i, 0] - b
        x = p1[i, 0] - x_mean
        y = p1[i, 1] - y_mean
        if val < 0:
            neg_x += x * x
            neg_y += y * y
        elif val > 0:
            pos_x += x * x
            pos_y += y * y
    cdef double neg_dist = sqrt(neg_x) + sqrt(neg_y)
    cdef double pos_dist = sqrt(pos_x) + sqrt(pos_y)
    x = abs(v2[0])
    y = abs(v2[1])
    if a < 0:
        if neg_dist > pos_dist:
            return -x, -y
        else:
            return x, y
    else:
        if neg_dist > pos_dist:
            return x, -y
        else:
            return -x, y


cdef double rotation_angle(double[:, :] p1, double x_mean, double y_mean) nogil:
    """Calculate rotation angle."""
    # Calculate the principal component axes (eigenvectors)
    cdef double cxx = 0
    cdef double cyy = 0
    cdef double cxy = 0
    cdef int i
    cdef double x, y
    for i in range(len(p1)):
        x = p1[i, 0] - x_mean
        y = p1[i, 1] - y_mean
        cxx += x * x
        cyy += y * y
        cxy += x * y
    cxx /= len(p1)
    cyy /= len(p1)
    cxy /= len(p1)
    cdef double[:, :] v
    with gil:
        _, v = eig(array([[cxx, cxy], [cxy, cyy]], dtype=f64))
    # Calculate the orientation of the axes
    x, y = axes_v(v[:, 1], v[:, 0], p1, x_mean, y_mean)
    cdef double a1 = atan2(y, x)
    x, y = axes_v(v[:, 0], v[:, 1], p1, x_mean, y_mean)
    cdef double a2 = atan2(y, x)
    # Calculate the rotation matrix
    if a1 * a2 > 0:
        return min(a1, a2)
    elif a1 * a2 < 0:
        if abs(a1) < M_PI / 2:
            return min(a1, a2)
        else:
            return max(a1, a2)
    else:
        return -M_PI / 2 if -M_PI / 2 in {a1, a2} else 0


cdef double[:, :] rotate(double[:, :] p1, double a) nogil:
    cdef double c = cos(a)
    cdef double s = sin(a)
    with gil:
        return array(p1) @ array([[c, -s], [s, c]], dtype=f64)


def norm_pca(path):
    """Normalization function by PCA."""
    cdef double[:, :] path_m = array(path, dtype=f64)
    _norm_pca(path_m)
    return array(path_m)


cdef void _norm_pca(double[:, :] p1) nogil:
    """Normalization implementation."""
    cdef double x_mean = 0
    cdef double y_mean = 0
    cdef int i
    for i in range(len(p1)):
        x_mean += p1[i, 0]
        y_mean += p1[i, 1]
    x_mean /= len(p1)
    y_mean /= len(p1)
    cdef double alpha = rotation_angle(p1, x_mean, y_mean)
    # Normalized the path points
    for i in range(len(p1)):
        p1[i, 0] -= x_mean
        p1[i, 1] -= y_mean
    p1[:] = rotate(p1, alpha)
    cdef int ind = 0
    cdef double p1_max = -INF
    cdef double p1_min = INF
    cdef double d_min = INF
    cdef double d
    for i in range(len(p1)):
        d = hypot(p1[i, 0], p1[i, 1])
        if d < d_min:
            d_min = d
        if p1[i, 0] > p1_max:
            p1_max = p1[i, 0]
            if p1[i, 1] > 0:
                ind = i
        if p1[i, 0] < p1_min:
            p1_min = p1[i, 0]
    d_min = p1_max - p1_min
    for i in range(len(p1)):
        p1[i, 0] /= d_min
        p1[i, 1] /= d_min
    # Swap to the starting point of the path
    roll(p1, ind)


cdef void transform(double[:, :] target) nogil:
    """Apply normalization and FFT to target."""
    _norm_pca(target)
    cdef double[:] real, imag
    with gil:
        c = array(target, dtype=f64)
        c = c[:, 0] + c[:, 1] * 1j
        v = fft(hstack([c, c, c]))[len(target):len(target) * 2]
        real = v.real
        imag = v.imag
    target[:, 0] = real
    target[:, 1] = imag


cdef double trapezoidal_camp(double[:] a, double[:] b) nogil:
    """Error comparison by trapezoidal rule."""
    cdef double area = 0
    cdef int i
    cdef double right, left, total
    for i in range(1, len(a)):
        right = a[i] - b[i]
        left = a[i - 1] - b[i - 1]
        total = abs(right) + abs(left)
        if right * left < 0:
            area += (right * right + left * left) / total * 0.5
        else:
            area += total * 0.5
    return area


@cython.final
cdef class NPlanar(ObjFunc):
    """A normalized matching method.

    Defects free. Normalized parameters are $[L_0, L_2, L_3, L_4, \\alpha]$.

    ![pxy](img/uniform_four_bar.png)
    """
    cdef int len
    cdef double[:, :] target

    def __cinit__(self, dict mech):
        self.target = array(mech['target'], dtype=f64)
        transform(self.target)
        self.len = len(self.target)
        self.lb = array([1e-5] * 4 + [0.], dtype=f64)
        self.ub = array([5.] * 4 + [2. * M_PI], dtype=f64)

    cdef double fitness(self, double[:] v) nogil:
        """Generate linkage with 5 parameters."""
        cdef double[:, :] p = c_uniform_path(v[None, :], self.len)[0]
        # NAN check
        cdef int i
        for i in range(len(p)):
            if isnan(p[i, 0]) or isnan(p[i, 1]):
                return HUGE_VAL
        transform(p)
        return (trapezoidal_camp(self.target[:, 0], p[:, 0]) +
                trapezoidal_camp(self.target[:, 1], p[:, 1]))

    cpdef object path(self, double[:] v):
        return array(c_uniform_path(v[None, :], self.len)[0])

    cpdef object result(self, double[:] v):
        return "M[" + ", ".join([(<VPoint> vp).expr()
                                 for vp in uniform_expr(v)]) + "]"



# -*- coding: utf-8 -*-

from typing import Callable, Optional
from enum import auto, IntEnum
from .utility import Algorithm, ObjFunc, FVal
from .config_types import DESetting

class Strategy(IntEnum):
    """Differential Evolution strategy."""
    S1 = auto()
    S2 = auto()
    S3 = auto()
    S4 = auto()
    S5 = auto()
    S6 = auto()
    S7 = auto()
    S8 = auto()
    S9 = auto()
    S0 = auto()

class DE(Algorithm):

    def __init__(
        self,
        func: ObjFunc[FVal],
        settings: DESetting,
        progress_fun: Optional[Callable[[int, str], None]] = None,
        interrupt_fun: Optional[Callable[[], bool]] = None
    ):
        """The argument `func` is a object inherit from [Verification],
        and all abstract methods should be implemented.

        The format of argument `settings`:

        + `strategy`: Strategy
            + type: int (0~9)
            + default: 0
        + `pop_num`: Population
            + type: int
            + default: 400
        + `F`: Weight factor
            + type: float (0.~1.)
            + default: 0.6
        + `CR`: Crossover rate
            + type: float (0.~1.)
            + default: 0.9
        + `max_gen` or `min_fit` or `max_time` or `slow_down`: Limitation of termination
            + type: int / float / float / float
            + default: Raise `ValueError`
        + `report`: Report per generation
            + type: int
            + default: 10

        !!! note
            The option `slow_down` is a percent value that
            current fitness difference of two generation is divide by last one.

        The argument `progress_fun` will be called when update progress,
        and the argument `interrupt_fun` will check the interrupt status from GUI or subprocess.
        """
        ...



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Differential Evolution

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from numpy import zeros, uint32 as usize, float64 as f64
from .utility cimport uint, rand_v, rand_i, ObjFunc, Algorithm


cpdef enum Strategy:
    S1
    S2
    S3
    S4
    S5
    S6
    S7
    S8
    S9
    S10


@cython.final
cdef class DE(Algorithm):
    """The implementation of Differential Evolution."""
    cdef Strategy strategy
    cdef double f, cr
    cdef uint[:] v
    cdef double[:] tmp
    cdef (double (*)(DE, uint) nogil) formula
    cdef (void (*)(DE, uint) nogil) setter

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        # strategy 0~9, choice what strategy to generate new member in temporary
        self.strategy = Strategy(settings['strategy'])
        # weight factor f is usually between 0.5 and 1 (in rare cases > 1)
        self.f = settings['f']
        if 0.5 > self.f or self.f > 1:
            raise ValueError('cr should be [0.5,1]')
        # crossover possible cr in [0,1]
        self.cr = settings['cr']
        if 0 > self.cr or self.cr > 1:
            raise ValueError('cr should be [0,1]')
        # the vector
        cdef uint num
        if self.strategy in {S1, S3, S6, S8}:
            num = 2
        elif self.strategy in {S2, S7}:
            num = 3
        elif self.strategy in {S4, S9}:
            num = 4
        else:
            num = 5
        self.v = zeros(num, dtype=usize)
        self.tmp = zeros(self.dim, dtype=f64)
        if self.strategy in {S1, S6}:
            self.formula = DE.f1
        elif self.strategy in {S2, S7}:
            self.formula = DE.f2
        elif self.strategy in {S3, S8}:
            self.formula = DE.f3
        elif self.strategy in {S4, S9}:
            self.formula = DE.f4
        else:
            self.formula = DE.f5
        if self.strategy in {S1, S2, S3, S4, S5}:
            self.setter = DE.s1
        else:
            self.setter = DE.s2

    cdef inline void init(self) nogil:
        """Initial population."""
        self.init_pop()
        self.find_best()

    cdef inline void vector(self, uint i) nogil:
        """Generate new vectors."""
        cdef uint j
        for j in range(len(self.v)):
            self.v[j] = i
            while True:
                if self.v[j] != i:
                    for k in range(j):
                        if self.v[j] == self.v[k]:
                            break
                    else:
                        break
                self.v[j] = rand_i(self.pop_num)

    cdef double f1(self, uint n) nogil:
        return self.best[n] + self.f * (
            self.pool[self.v[0], n] - self.pool[self.v[1], n])

    cdef double f2(self, uint n) nogil:
        return self.pool[self.v[0], n] + self.f * (
            self.pool[self.v[1], n] - self.pool[self.v[2], n])

    cdef double f3(self, uint n) nogil:
        return self.tmp[n] + self.f * (self.best[n] - self.tmp[n]
            + self.pool[self.v[0], n] - self.pool[self.v[1], n])

    cdef double f4(self, uint n) nogil:
        return self.best[n] + self.f * (
            self.pool[self.v[0], n] + self.pool[self.v[1], n]
            - self.pool[self.v[2], n] - self.pool[self.v[3], n])

    cdef double f5(self, uint n) nogil:
        return self.pool[self.v[4], n] + self.f * (
            self.pool[self.v[0], n] + self.pool[self.v[1], n]
            - self.pool[self.v[2], n] - self.pool[self.v[3], n])

    cdef void s1(self, uint n) nogil:
        for _ in range(self.dim):
            self.tmp[n] = self.formula(self, n)
            n = (n + 1) % self.dim
            if rand_v() >= self.cr:
                break

    cdef void s2(self, uint n) nogil:
        cdef uint l_v
        for l_v in range(self.dim):
            if rand_v() < self.cr or l_v == self.dim - 1:
                self.tmp[n] = self.formula(self, n)
            n = (n + 1) % self.dim

    cdef inline void recombination(self, int i) nogil:
        """use new vector, recombination the new one member to tmp."""
        self.tmp[:] = self.pool[i, :]
        self.setter(self, rand_i(self.dim))

    cdef inline void generation(self) nogil:
        cdef uint i, s
        cdef double tmp_f
        for i in range(self.pop_num):
            # Generate a new vector
            self.vector(i)
            # Use the vector recombine the member to temporary
            self.recombination(i)
            # Check the one is out of bound
            for s in range(self.dim):
                if not self.func.ub[s] >= self.tmp[s] >= self.func.lb[s]:
                    break
            else:
                # Test
                tmp_f = self.func.fitness(self.tmp)
                # Self evolution
                if tmp_f < self.fitness[i]:
                    self.assign_from(i, tmp_f, self.tmp)
        self.find_best()



# -*- coding: utf-8 -*-

"""Pyslvs graph functions."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from .graph import (
    link_assortment, contracted_link_assortment, labeled_enumerate, Graph,
)
from .planar import is_planar
from .layout import external_loop_layout
from .structural import (
    conventional_graph, contracted_graph, link_synthesis,
    contracted_link_synthesis,
)



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Fast planar linkage synthesis. (defect allowable)

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from collections import OrderedDict
from logging import getLogger
from numpy import (
    zeros, array, arange, interp, argmax, concatenate, float64 as f64,
)
from numpy.core.multiarray import correlate
from libc.math cimport (
    cos, sin, atan2, isnan, log, INFINITY as INF, HUGE_VAL, M_PI,
)
from libcpp.pair cimport pair
from libcpp.map cimport map
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from pyslvs.expression cimport Coord, VPoint, distance
from pyslvs.metaheuristics.utility cimport ObjFunc
from pyslvs.tinycadlib cimport (
    quick_solve, I_LABEL, A_LABEL, P_LABEL, PLA, PLAP, PLLP, preprocessing,
    CCoord, Sym, SwappablePair,
)
from pyslvs.topo_config cimport t_config, EStack, Expr
from pyslvs.bfgs cimport SolverSystem

try:
    from scipy.signal import fftconvolve
except ImportError:
    fftconvolve = None

logger = getLogger()


def norm_path(path, scale=1):
    """Normalization function."""
    cdef double[:, :] path_m = array(path, dtype=f64)
    _norm(path_m, scale)
    return array(path_m)


cdef void _norm(double[:, :] path, double scale) nogil:
    """Normalization implementation."""
    cdef CCoord centre = CCoord(0, 0)
    cdef double x, y
    cdef int i
    for i in range(len(path)):
        centre.x += path[i, 0]
        centre.y += path[i, 1]
    cdef int end = len(path)
    centre.x /= end
    centre.y /= end
    cdef vector[double] angle = vector[double](end + 1)
    cdef vector[double] length = vector[double](end + 1)
    cdef int sp = 0
    for i in range(len(path)):
        x = path[i, 0]
        y = path[i, 1]
        angle[i] = atan2(y - centre.y, x - centre.x)
        length[i] = distance(centre.x, centre.y, x, y)
        if length[i] > length[end]:
            length[end] = length[i]
            angle[end] = angle[i]
            sp = end
    roll(path, sp)
    cdef CCoord bound = CCoord(INF, -INF)
    cdef double a
    for i in range(len(path)):
        a = angle[i] - angle[end]
        path[i, 0] = length[i] * cos(a)
        path[i, 1] = length[i] * sin(a)
        if path[i, 0] < bound.x:
            bound.x = path[i, 0]
        if path[i, 0] > bound.y:
            bound.y = path[i, 0]
    scale /= (bound.y - bound.x)
    _mul1d(path[:, 0], scale)
    _mul1d(path[:, 1], scale)


cdef void roll(double[:, :] path, int ind) nogil:
    """Split 1D path from sp, concatenate to end."""
    if ind == 0:
        return
    cdef double[:, :] tmp
    with gil:
        tmp = zeros((ind, 2), dtype=f64)
    tmp[:] = path[:ind]
    path[:len(path) - ind] = path[ind:]
    path[len(path) - ind:] = tmp[:]


def curvature(path):
    r"""Calculate the signed curvature and return as an array.

    $$
    \kappa(t) = \frac{x'y'' - x''y'}{(x'^2 + y'^2)^\frac{3}{2}}
    $$
    """
    cdef double[:, :] path_m = array(path, dtype=f64)
    return array(_curvature(path_m))


cdef double[:] _curvature(double[:, :] path) nogil:
    """Calculate the signed curvature."""
    cdef double[:, :] p1d = _derivative(path)
    cdef double[:, :] p2d = _derivative(p1d)
    cdef double[:] k
    with gil:
        k = zeros(len(path) - 2, dtype=f64)
    cdef int i
    for i in range(len(path) - 2):
        k[i] = ((p1d[i, 0] * p2d[i, 1] - p2d[i, 0] * p1d[i, 1])
                / (p1d[i, 0] * p1d[i, 0] + p1d[i, 1] * p1d[i, 1]) ** 1.5)
    return k


def derivative(double[:, :] p):
    """Differential function. Return $p'$."""
    return array(_derivative(p))


cdef double[:, :] _derivative(double[:, :] p) nogil:
    """Differential function backend."""
    cdef double[:, :] pd
    with gil:
        pd = zeros((len(p), 2), dtype=f64)
    cdef double max0 = 0
    cdef double max1 = 0
    cdef int i, j
    for i in range(len(p)):
        j = i + 1
        if j >= len(p):
            j = 0
        pd[i, 0] = p[j, 0] - p[i, 0]
        pd[i, 1] = p[j, 1] - p[i, 1]
        if pd[i, 0] > max0:
            max0 = pd[i, 0]
        if pd[i, 1] > max1:
            max1 = pd[i, 1]
    i = len(p) - 1
    if max0 == pd[i, 0] or max1 == pd[i, 1]:
        return pd[:i]
    else:
        return pd


def path_signature(double[:] k, double maximum = 100):
    r"""Require a curvature, return path signature.
    It's composed by curvature $\kappa$ and a $K$ value.

    $$
    K = \int^t_0 |\kappa(t)| dt
    $$

    >>> from pyslvs.optimization import curvature, path_signature
    >>> path_signature(curvature(...))
    """
    return array(_path_signature(k, maximum))


cdef double[:, :] _path_signature(double[:] k, double maximum) nogil:
    """Require a curvature, return path signature."""
    cdef double[:, :] s
    with gil:
        s = zeros((len(k), 2), dtype=f64)
    cdef double interval = maximum / len(k)
    cdef double v = 0
    cdef int i
    for i in range(len(k)):
        s[i] = v
        v += interval
    s[:, 1] = k
    return s


def cross_correlation(double[:, :] p1, double[:, :] p2, double t = 0.1):
    r"""Compare signature and return as an 1d array.

    $$
    \begin{aligned}
    C_n(j, W, P) &= \left|\sum_i^{l_P} \frac{(W_{i + j}
    - \overline{W}_{j\rightarrow j + l_P})(P_i-\overline{P})}{
    \sqrt{\sum_i^{l_P}(W_{i + j} - \overline{W}_{j\rightarrow j + l_P})^2
    \sum_i^{l_P}(P_i - \overline{P})^2}}\right|
    \\
    S &= \arg\max\{C_n(j)\} t
    \end{aligned}
    $$

    >>> from pyslvs.optimization import curvature, path_signature
    >>> ps1 = path_signature(curvature(...))
    >>> ps2 = path_signature(curvature(...))
    >>> from pyslvs.optimization import cross_correlation
    >>> cc = cross_correlation(ps1, ps2)
    """
    return array(_cross_correlation(p1, p2, t), dtype=f64)


cdef double[:] _cross_correlation(double[:, :] ps1, double[:, :] ps2, double t):
    """Compare path signature."""
    cdef double[:] p1 = interp(arange(0, _extr1d(ps1[:, 0], 1), t),
                               ps1[:, 0], ps1[:, 1])
    cdef double[:] p2 = interp(arange(0, _extr1d(ps2[:, 0], 1), t),
                               ps2[:, 0], ps2[:, 1])
    p1 = concatenate((p1, p1))
    cdef int s1 = len(p1)
    cdef int s2 = len(p2)
    cdef int n = s1 + s2 - 1
    if s2 >= s1:
        s1, s2 = s2, s1
    if (fftconvolve is not None and 1.89095737e-9 * 3 * n * log(n)
        < 2.1364985e-10 * (s1 - s2 + 1) * s2 - 1e-3
    ):
        return fftconvolve(p1, p2[::-1], mode='valid')
    else:
        return correlate(p1, p2)


cdef double _extr1d(double[:] s, bint op) nogil:
    """Max value of 1D slice."""
    cdef double m = -INF if op else INF
    for v in s:
        if op:
            if v > m:
                m = v
        elif v < m:
            m = v
    return m


cdef void _mul1d(double[:] s, double v) nogil:
    """Inplace multiplication assignment of 1D slice."""
    cdef int i
    for i in range(len(s)):
        s[i] *= v


cdef double[:, :] _slice_nan2d(double[:, :] s) nogil:
    """Slice continuous view without NaN."""
    cdef int first = -1
    cdef int second = -1
    cdef int i
    cdef double v
    for i in range(len(s)):
        v = s[i, 1]
        if i == first + 1 and isnan(v):
            # The first of slice
            first = i
        elif isnan(v):
            # The end of slice
            second = i
            break
    if first == -1:
        first = 0
    if second == -1:
        second = len(s) - 1
    return s[first:second]


@cython.final
cdef class FPlanar(ObjFunc):
    """A fast matching method that adds mapping angles to variables.

    Allowing defects.
    """
    cdef bint bfgs_mode, shape_only, use_curvature, full_path, ordered
    cdef int target_count, target_len, input_count, l_base
    cdef public unsigned long callback
    cdef list vpoints
    cdef long[:] target_nodes, pivots
    cdef double[:, :, :] target
    cdef EStack exprs
    cdef cset[int] slider
    cdef map[Sym, CCoord] joint_pos
    cdef map[Sym, double] param

    def __cinit__(self, dict mech):
        # mech = {
        #     'expression': List[VPoint],
        #     'input': OrderedDict([((b0, d0), [start, end]), ...]),
        #     'placement': {pt: (x, y, r)},
        #     'target': {pt: [(x0, y0), (x1, y1), ...]},
        #     'same': {pt: match_to_pt},
        #     # Bounds of base link length
        #     'upper': float,
        #     'lower': float,
        #     'shape_only': bool,
        #     'use_curvature': bool,
        #     'full_path': bool,
        # }
        placement = mech.get('placement', {})
        if len(placement) == 0:
            raise ValueError("no grounded joint")
        target = mech.get('target', {})
        if len(target) == 0:
            raise ValueError("no target joint")
        check_set = {len(t) for t in target.values()}
        if len(check_set) != 1:
            raise ValueError("target paths should be in the same size")
        self.target_len = check_set.pop()
        # Change the target paths into memory view
        self.target_count = len(target)
        self.target_nodes = zeros(self.target_count, dtype=int)
        self.target = zeros((self.target_count, self.target_len, 2), dtype=f64)
        same = mech.get('same', {})
        self.shape_only = mech.get('shape_only', False)
        self.use_curvature = mech.get('use_curvature', False)
        self.full_path = mech.get('full_path', False)
        if self.use_curvature:
            self.shape_only = False
        cdef int i, j
        cdef double[:, :] path
        for i, j in enumerate(target):
            self.target_nodes[i] = j
            path = array(target[j], dtype=f64)
            if self.shape_only:
                _norm(path, 1)
            if self.use_curvature:
                path = _path_signature(_curvature(path), 100)
            self.target[i, :, :] = path
        # Expressions (must be readonly)
        self.vpoints = list(mech.get('expression', []))
        self.pivots = array([i for i, vp in enumerate(self.vpoints)
                             if (<VPoint>vp).grounded()], dtype=int)
        self.slider = {i for i, vp in enumerate(self.vpoints)
                       if (<VPoint>vp).is_slider()}
        inputs = OrderedDict(mech.get('input', {}))
        self.input_count = len(inputs)
        status = {}
        self.exprs = t_config(self.vpoints, tuple(inputs.keys()), status)
        self.bfgs_mode = not all(status.values())
        if not preprocessing(self.exprs, self.vpoints, {p: 0. for p in inputs},
                             self.joint_pos, map[SwappablePair, double](),
                             self.param):
            raise ValueError("wrong number of input parameters")
        # Bounds
        ub = []
        lb = []
        # Position
        cdef double x, y, r
        for i in self.pivots:
            x, y, r = placement[i]
            ub.append(x + r)
            ub.append(y + r)
            lb.append(x - r)
            lb.append(y - r)
        # Length of links
        link_upper = float(mech.get('upper', 100))
        link_lower = float(mech.get('lower', 0))
        cdef Expr expr
        for expr in self.exprs.stack:
            ub.append(link_upper)
            lb.append(link_lower)
            if expr.func in {PLA, PLAP} and expr.v2.first == A_LABEL:
                # The included angle of the link
                ub.append(2 * M_PI)
                lb.append(0)
            elif expr.func == PLLP:
                ub.append(link_upper)
                lb.append(link_lower)
        # The start of the angle parameters
        self.l_base = len(ub)
        if self.use_curvature and self.full_path:
            # Scale factor
            ub.append(1)
            lb.append(1e-12)
        else:
            # Input nodes
            for start, end in inputs.values():
                ub.append(start / 180 * M_PI)
                lb.append(end / 180 * M_PI)
            # Angle rage (input count * target count)
            ub[self.l_base:] *= self.target_len
            lb[self.l_base:] *= self.target_len
        self.ub = array(ub, dtype=f64)
        self.lb = array(lb, dtype=f64)
        # Swap upper and lower bound if reversed
        for i in range(len(self.ub)):
            if self.ub[i] < self.lb[i]:
                self.ub[i], self.lb[i] = self.lb[i], self.ub[i]
        self.callback = 0

    cpdef bint is_two_kernel(self):
        """Input a generic data (variable array), return the mechanism
        expression.
        """
        return self.bfgs_mode

    cdef double fitness(self, double[:] v) nogil:
        """Return the difference of the path signature.

        + Position of fix joints.
        + Link lengths.
        + Angle corresponding to the target points.
        """
        self.callback += 1
        cdef map[int, vector[CCoord]] target
        cdef map[Sym, double] param = map[Sym, double](self.param)
        cdef map[Sym, CCoord] joint_pos = map[Sym, CCoord](self.joint_pos)
        cdef bint ok
        cdef int i, j, node, vi
        cdef double x, y
        cdef Expr expr
        cdef CCoord c
        cdef pair[Sym, CCoord] jp
        for i in range(self.target_len):
            vi = 0
            for node in self.pivots:
                joint_pos[Sym(P_LABEL, node)] = CCoord(v[vi], v[vi + 1])
                vi += 2
            # Input parameters (length)
            for expr in self.exprs.stack:
                param[expr.v1] = v[vi]
                vi += 1
                if expr.v2.first == I_LABEL:
                    continue
                param[expr.v2] = v[vi]
                vi += 1
            # Input parameters (angles)
            for vi in range(self.input_count):
                param[Sym(I_LABEL, vi)] = v[self.l_base + vi
                                            + i * self.target_count]
            # Solve
            ok, joint_pos = quick_solve(self.exprs.stack, joint_pos, param)
            if not ok:
                return HUGE_VAL
            if self.bfgs_mode:
                with gil:
                    data_dict = {}
                    for jp in joint_pos:
                        data_dict[jp.first.second] = Coord.__new__(Coord,
                                                                   jp.second.x,
                                                                   jp.second.y)
                    # Solve
                    try:
                        solved_bfgs = SolverSystem(self.vpoints, {}, data_dict).solve()
                    except ValueError:
                        return HUGE_VAL
            # Collecting
            for node in self.target_nodes:
                if self.bfgs_mode:
                    with gil:
                        x, y = solved_bfgs[node][0]
                    target[node].push_back(CCoord(x, y))
                else:
                    c = joint_pos[Sym(P_LABEL, node)]
                    target[node].push_back(c)
        # Compare
        cdef double fitness = 0
        cdef double scale
        cdef double[:, :] path1, path2
        for vi, node in enumerate(self.target_nodes):
            if self.use_curvature:
                with gil:
                    path1 = zeros((self.target_len, 2), dtype=f64)
                    path2 = array(self.target[vi], dtype=f64)
                for i in range(self.target_len):
                    c = target[node][i]
                    path1[i, 0] = c.x
                    path1[i, 1] = c.y
                path1 = _slice_nan2d(path1)
                if len(path1) == 0:
                    return HUGE_VAL
                path1 = _path_signature(_curvature(path1), 100)
                scale = 1 / v[len(v) - 1]
                if not self.full_path:
                    scale *= _extr1d(path1[:, 0], 1) / _extr1d(path2[:, 0], 1)
                _mul1d(path2[:, 0], scale)
                with gil:
                    j = argmax(_cross_correlation(path2, path1, 0.1))
                for i in range(len(path2)):
                    path2[i, 0] += j
                for i in range(self.target_len):
                    fitness += path1[i, 0] - path2[i, 0]
            else:
                if self.shape_only:
                    with gil:
                        path1 = zeros((self.target_len, 2), dtype=f64)
                    for i in range(self.target_len):
                        c = target[node][i]
                        path1[i, 0] = c.x
                        path1[i, 1] = c.y
                    _norm(path1, 1)
                    for i in range(self.target_len):
                        target[node][i].x = path1[i, 0]
                        target[node][i].y = path1[i, 1]
                for i in range(self.target_len):
                    c = target[node][i]
                    fitness += distance(c.x, c.y,
                                        self.target[vi, i, 0],
                                        self.target[vi, i, 1])
        return fitness

    cpdef object result(self, double[:] v):
        """Input a generic data (variable array), return the mechanism
        expression.
        """
        cdef map[Sym, double] param = map[Sym, double](self.param)
        cdef map[Sym, CCoord] joint_pos = map[Sym, CCoord](self.joint_pos)
        cdef int vi = 0
        cdef int node
        for node in self.pivots:
            joint_pos[Sym(P_LABEL, node)] = CCoord(v[vi], v[vi + 1])
            vi += 2
        cdef Expr expr
        for expr in self.exprs.stack:
            param[expr.v1] = v[vi]
            vi += 1
            if expr.v2.first == I_LABEL:
                continue
            param[expr.v2] = v[vi]
            vi += 1
        # Input parameters (angles)
        for vi in range(self.input_count):
            param[Sym(I_LABEL, vi)] = v[self.l_base + vi]
        # Solve
        _, joint_pos = quick_solve(self.exprs.stack, joint_pos, param)
        cdef pair[Sym, CCoord] jp
        if self.bfgs_mode:
            data_dict = {}
            for jp in joint_pos:
                data_dict[jp.first.second] = Coord.__new__(Coord,
                                                           jp.second.x,
                                                           jp.second.y)
            # Solve
            solved_bfgs = SolverSystem(self.vpoints, {}, data_dict).solve()
        # Collecting
        cdef CCoord c
        cdef VPoint vp
        cdef double x, y
        for node in range(len(self.vpoints)):
            vp = self.vpoints[node]
            if self.bfgs_mode:
                x, y = solved_bfgs[node][0]
                vp.locate(x, y)
                if vp.is_slider():
                    vp.move(solved_bfgs[node][0], solved_bfgs[node][1])
            else:
                c = joint_pos[Sym(P_LABEL, node)]
                if vp.is_slider():
                    vp.locate(vp.c[0, 0], vp.c[0, 1])
                    vp.move((vp.c[0, 0], vp.c[0, 1]), (c.x, c.y))
                else:
                    vp.locate(c.x, c.y)
        expr_str = "M[" + ", ".join([(<VPoint> vp).expr()
                                     for vp in self.vpoints]) + "]"
        logger.debug(f"Result: {expr_str}")
        return expr_str



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Particle Swarm Optimization

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from numpy import zeros, float64 as f64
from .utility cimport uint, rand_v, ObjFunc, Algorithm


@cython.final
cdef class PSO(Algorithm):
    cdef double cognition, social, velocity
    cdef double[:] best_f_past
    cdef double[:, :] best_past

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        self.cognition = settings['cognition']
        self.social = settings['social']
        self.velocity = settings['velocity']
        self.best_past = zeros((self.pop_num, self.dim), dtype=f64)
        self.best_f_past = zeros(self.pop_num, dtype=f64)

    cdef inline void init(self) nogil:
        """Initial population."""
        self.best_past[:, :] = self.pool
        self.best_f_past[:] = self.fitness

    cdef inline void generation(self) nogil:
        cdef double alpha
        cdef double beta
        cdef uint i, s
        for i in range(self.pop_num):
            alpha = self.cognition * rand_v()
            beta = self.social * rand_v()
            for s in range(self.dim):
                self.pool[i, s] = self.check(s, self.velocity * self.pool[i, s]
                    + alpha * (self.best_past[i, s] - self.pool[i, s])
                    + beta * (self.best[s] - self.pool[i, s]))
            self.fitness[i] = self.func.fitness(self.pool[i, :])
            if self.fitness[i] < self.best_f_past[i]:
                self.best_past[i, :] = self.pool[i, :]
                self.best_f_past[i] = self.fitness[i]
            if self.fitness[i] < self.best_f:
                self.set_best(i)



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Triangular expressions.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from libcpp.pair cimport pair
from libcpp.vector cimport vector
from .tinycadlib cimport Expr

ctypedef pair[int, int] Sym

cpdef EStack t_config(object vpoints_, object inputs_, object status_=*)

cdef class EStack:
    cdef vector[Expr] stack

    cdef void add_pxy(self, Sym c1, Sym v1, Sym v2, Sym t) nogil
    cdef void add_ppp(self, Sym c1, Sym c2, Sym c3, Sym t) nogil
    cdef void add_pla(self, Sym c1, Sym v1, Sym v2, Sym t) nogil
    cdef void add_plap(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t) nogil
    cdef void add_pllp(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t) nogil
    cdef void add_plpp(self, Sym c1, Sym v1, Sym c2, Sym c3, Sym t, bint op) nogil
    cdef void add_palp(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t, bint op) nogil
    cpdef list as_list(self)



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Graph class.

The algorithm reference:
+ NetworkX

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libcpp.pair cimport pair
import sys
from typing import Tuple, Dict, Iterator
from itertools import permutations, groupby
from numpy import zeros, uint8 as np_uint

ctypedef pair[int, int] ipair


cpdef list link_assortment(Graph g):
    """Return link assortment of the graph."""
    assortment = [0]
    if not g.edges:
        return assortment
    cdef int d, n
    for n in g.degrees().values():
        if n < 2:
            continue
        d = n - 2
        while d >= len(assortment):
            assortment.append(0)
        assortment[d] += 1
    return assortment


cpdef list contracted_link_assortment(Graph g):
    """Return contracted link assortment of the graph."""
    if not g.edges:
        return [0]
    assortment = [0] * link_assortment(g)[0]
    counted = set()
    for mcl in _multi_contracted_links(g, False):
        counted.update(mcl)
        assortment[len(mcl) - 1] += 1
    # For single contracted links
    cdef int n, d
    for n, d in g.degrees().items():
        if d != 2:
            continue
        if n not in counted:
            assortment[0] += 1
    return assortment


cdef list _multi_contracted_links(Graph g, bint only_one):
    """Return a list of multiple contracted links."""
    cdef int n1, n2, index, neighbor
    contracted_links = []
    counted = set()
    for n1 in g.vertices:
        # Only for binary link.
        if len(g.adj[n1]) != 2:
            continue
        # n1 is not collected yet.
        if n1 in counted:
            continue
        index = 0
        c_links = [n1]
        while index < len(c_links):
            for neighbor in g.adj[c_links[index]]:
                if len(g.adj[neighbor]) == 2 and neighbor not in c_links:
                    c_links.append(neighbor)
            index += 1
        if len(c_links) > 1:
            if only_one:
                return c_links
            counted.update(c_links)
            contracted_links.append(tuple(c_links))
    if only_one:
        return []
    else:
        return contracted_links


cpdef list labeled_enumerate(Graph g):
    """Enumerate each node with labeled except isomorphism."""
    result = []
    cdef int n1, n2
    cdef Graph g1, g2
    for n1 in g.vertices:
        g1 = Graph.__new__(Graph, [e for e in g.edges if n1 not in e])
        for n2, g2 in result:
            if g1.is_isomorphic(g2):
                break
        else:
            result.append((n1, g1))
    return result


@cython.final
cdef class Graph:
    """The undirected graph class, support multigraph."""

    def __cinit__(self, object edges):
        self.edges = tuple(edges)
        # vertices
        cdef int p1, p2
        vertices = []
        for p1, p2 in self.edges:
            if p1 not in vertices:
                vertices.append(p1)
            if p2 not in vertices:
                vertices.append(p2)
        self.vertices = tuple(vertices)
        # adj
        self.adj = {n: self.neighbors(n) for n in self.vertices}

    cpdef void add_vertices(self, object vertices):
        """Add vertices from iterable object `vertices`."""
        self.vertices = tuple(set(self.vertices) | set(vertices))

    cpdef void add_edge(self, int n1, int n2):
        """Add edge `n1` to `n2`."""
        self.edges += ((n1, n2),)
        self.vertices = tuple(set(self.vertices) | {n1, n2})
        cdef int n
        for n in (n1, n2):
            self.adj[n] = self.neighbors(n)

    cpdef void add_path(self, object new_nodes):
        edges = list(self.edges)
        vertices = set(self.vertices)
        cdef int n1 = -1
        cdef int n2
        for n2 in new_nodes:
            if n1 != -1:
                edges.append((n1, n2))
                vertices.update((n1, n2))
            n1 = n2
        self.edges = tuple(edges)
        self.vertices = tuple(vertices)
        for n1 in vertices:
            self.adj[n1] = self.neighbors(n1)

    cpdef void remove_edge(self, int n1, int n2):
        """Remove edge(s) {n1, n2} once if exist, otherwise do nothing."""
        vertices = set()
        edges = []
        cdef bint once = False
        for edge in self.edges:
            if {n1, n2} != set(edge) or once:
                vertices.update(edge)
                edges.append(edge)
            else:
                once = True
        self.edges = tuple(edges)
        self.vertices = tuple(vertices)
        cdef int n
        for n in self.vertices:
            edge = self.neighbors(n)
            if edge:
                self.adj[n] = edge
            else:
                self.adj.pop(n, None)

    cpdef int dof(self):
        """Return DOF of the graph.

        !!! note
            DOF is the Degree of Freedoms to a mechanism.

            In the [Graph] objects, all vertices will assumed as revolute
            joints (1 DOF).

            $$
            F = 3(N_L - 1) - 2N_J
            $$
        """
        return 3 * (len(self.vertices) - 1) - (2 * len(self.edges))

    cpdef inline tuple neighbors(self, int n):
        """Return the neighbors of the vertex `n`."""
        neighbors = []
        cdef int l1, l2
        for l1, l2 in self.edges:
            if n == l1:
                neighbors.append(l2)
            if n == l2:
                neighbors.append(l1)
        return tuple(neighbors)

    cpdef dict degrees(self):
        """Return the degrees of each vertex."""
        return {n: len(neighbors) for n, neighbors in self.adj.items()}

    cpdef ullong degree_code(self):
        """Generate a degree code.

        With a sorted vertices mapping by the degrees of each vertex,
        regenerate a new adjacency matrix.
        A binary code can be found by concatenating the upper right elements.
        The degree code is the maximum value of the permutation.
        """
        if len(self.vertices) < 2:
            return 0
        # Create a new mapping
        degrees = self.degrees()
        # Permute the group to find the max degree code
        # "per1" is the mapping of the graph
        # "prefix" is the last choose of the candidate (can be multiple)
        per1 = []
        prefix = []
        cdef int i, n1, n2
        cdef ullong code, sub_code
        for _, g in groupby(
            sorted(degrees, key=degrees.__getitem__, reverse=True),
            key=degrees.__getitem__
        ):
            code = 0
            order = None
            g = tuple(g)
            for pre in iter(prefix if len(prefix) > 1 else [()]):
                for per2 in permutations(g):
                    if len(per2) == 1:
                        order = pre + per2
                        prefix = [order]
                        break
                    sub_code = 0
                    # Calculate sub code
                    per3 = tuple(per1) + pre + per2
                    for i, n1 in enumerate(per3):
                        for n2 in per3[i + 1:]:
                            sub_code <<= 1
                            sub_code += n2 in self.adj[n1]
                    # Compare sub code
                    if sub_code > code or order is None:
                        code = sub_code
                        order = pre + per2
                        prefix = [order]
                    elif sub_code == code:
                        # Will affect the subsequent arrangement
                        prefix.append(pre + per2)
            if len(prefix) == 1:
                per1.extend(order)
        # Check the last one prefix candidate
        if not set(order) <= set(per1):
            per1.extend(order)
        # Calculate the degree code
        code = 0
        for i, n1 in enumerate(per1):
            for n2 in per1[i + 1:]:
                code <<= 1
                code += n2 in self.adj[n1]
        return code

    cpdef double[:, :] adjacency_matrix(self):
        """Generate a adjacency matrix.

        Assume the matrix $A[i, j] = A[j, i]$.
        Where $A[i, j] = 1$ if edge `(i, j)` exist.
        """
        cdef int n = len(self.vertices)
        cdef double[:, :] am = zeros((n, n), dtype=np_uint)
        cdef int n1, n2
        for n1, n2 in self.edges:
            am[n1, n2] += 1
            am[n2, n1] += 1
        return am

    cpdef bint is_connected(self, int without = -1):
        """Return `True` if the graph is connected.
        Set the argument `without` to ignore one vertex.
        """
        if not self.vertices:
            return True
        cdef int neighbors
        cdef int index = 0
        vertices = []
        # Change start node if index zero has been excluded.
        if without == self.vertices[0]:
            vertices.append(self.vertices[1])
        else:
            vertices.append(self.vertices[0])
        # Search node by node.
        while index < len(vertices):
            for neighbors in self.adj[vertices[index]]:
                if (neighbors not in vertices) and neighbors != without:
                    vertices.append(neighbors)
            index += 1
        if without != -1:
            vertices.append(without)
        return len(vertices) == len(self.vertices)

    cpdef bint has_cut_link(self):
        """Return true if the graph has any cut links."""
        cdef int n, d
        for n, d in self.degrees().items():
            # Only for multiple links.
            if d > 2:
                # Remove a multiple link should be keep connected.
                if not self.is_connected(n):
                    return True
        return False

    cpdef bint is_degenerate(self):
        """Return true if this kinematic chain is degenerate.

        + Prue all multiple contracted links recursively.
        + Check the DOF of sub-graph if it is lower then zero.
        """
        if self.has_triangle():
            return True
        cdef int n1, n2
        cdef Graph g = self.copy()
        mcl = set()
        while True:
            mcl.update(_multi_contracted_links(g, True))
            if not mcl:
                break
            c_l = []
            for n1, n2 in g.edges:
                if not {n1, n2} & mcl:
                    c_l.append((n1, n2))
            mcl.clear()
            # Pruned graph
            g = Graph.__new__(Graph, c_l)
            if {n2 for n2 in g.degrees().values()} == {2}:
                # The graph is a basic loop.
                break
        # Check the DOF
        return g.dof() < 1

    cpdef bint has_triangle(self):
        """Return true if the graph has triangle."""
        cdef int n1, n2
        for neighbors in self.adj.values():
            for n1 in neighbors:
                for n2 in neighbors:
                    if n1 == n2:
                        continue
                    if n1 in self.adj[n2]:
                        return True
        return False

    cpdef bint is_isomorphic(self, Graph g):
        """Return true if two graphs is isomorphic.

        Default is using VF2 algorithm.
        """
        return self.is_isomorphic_vf2(g)

    cpdef bint is_isomorphic_vf2(self, Graph g):
        """Compare isomorphism by VF2 algorithm,
        one of the high performance isomorphic algorithms.
        """
        cdef GraphMatcher gm_gh = GraphMatcher.__new__(GraphMatcher, self, g)
        return gm_gh.is_isomorphic()

    cpdef bint is_isomorphic_degree_code(self, Graph g):
        """Compare isomorphism by degree code algorithm.

        + <https://doi.org/10.1115/1.2919236>
        """
        return self.degree_code() == g.degree_code()

    cpdef Graph duplicate(self, object vertices, int times):
        """Make graph duplicate by specific `vertices`. Return a new graph."""
        if times < 1:
            raise ValueError("please input a number larger than 1.")
        cdef int max_num = max(self.vertices) + 1
        mapping = {}
        cdef int i, n1, n2
        edges = set(self.edges)
        for i in range(times):
            for n1 in sorted(set(vertices)):
                mapping[n1] = max_num
                max_num += 1
            for n1, n2 in self.edges:
                if n1 in mapping:
                    n1 = mapping[n1]
                if n2 in mapping:
                    n2 = mapping[n2]
                if n1 > n2:
                    n1, n2 = n2, n1
                edges.add((n1, n2))
        return Graph.__new__(Graph, edges)

    cpdef Graph copy(self):
        """The copy method of the Graph object."""
        return Graph.__new__(Graph, self.edges)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self.edges)})"


cdef bint _is_adjacent(Graph g, int u, int v):
    """Find the distance between u and v."""
    if v in g.adj[u]:
        return True
    return False


@cython.final
cdef class GraphMatcher:
    """GraphMatcher and GMState class from NetworkX.

    Copyright (C) 2007-2009 by the NetworkX maintainers
    All rights reserved.
    BSD license.

    This work was originally coded by Christopher Ellison
    as part of the Computational Mechanics Python (CMPy) project.
    James P. Crutchfield, principal investigator.
    Complexity Sciences Center and Physics Department, UC Davis.
    """
    cdef Graph g1, g2
    cdef set g1_nodes, g2_nodes
    cdef dict core_1, core_2, inout_1, inout_2, mapping
    cdef GMState state

    def __cinit__(self, Graph g1, Graph g2):
        self.g1 = g1
        self.g2 = g2
        self.g1_nodes = set(g1.vertices)
        self.g2_nodes = set(g2.vertices)
        # Set recursion limit
        cdef int old_recursion_limit = sys.getrecursionlimit()
        cdef int expected_max_recursion_level = len(self.g2.vertices)
        if old_recursion_limit < 1.5 * expected_max_recursion_level:
            # Give some breathing room
            sys.setrecursionlimit(int(1.5 * expected_max_recursion_level))
        # Initialize state
        self.initialize()

    cdef inline void initialize(self):
        """Re-initializes the state of the algorithm."""
        # core_1[n] contains the index of the node paired with n, which is m,
        #            provided n is in the mapping.
        # core_2[m] contains the index of the node paired with m, which is n,
        #            provided m is in the mapping.
        self.core_1 = {}
        self.core_2 = {}
        # See the paper for definitions of M_x and T_x^{y}
        # inout_1[n]  is non-zero if n is in M_1 or in T_1^{inout}
        # inout_2[m]  is non-zero if m is in M_2 or in T_2^{inout}
        #
        # The value stored is the depth of the SSR tree when the node became
        # part of the corresponding set.
        self.inout_1 = {}
        self.inout_2 = {}
        # Practically, these sets simply store the nodes in the subgraph.
        self.state = GMState.__new__(GMState, self)
        # Provide a convenient way to access the isomorphism mapping.
        self.mapping = self.core_1.copy()

    def candidate_pairs_iter(self) -> Iterator[Tuple[int, int]]:
        """Iterator over candidate pairs of nodes in g1 and g2."""
        # First we compute the inout-terminal sets.
        s1 = set(self.inout_1) - set(self.core_1)
        s2 = set(self.inout_2) - set(self.core_2)
        t1_inout = [node for node in self.g1_nodes if node in s1]
        t2_inout = [node for node in self.g2_nodes if node in s2]
        # If t1_inout and t2_inout are both nonempty.
        # P(s) = t1_inout x {min t2_inout}
        cdef int node
        if t1_inout and t2_inout:
            for node in t1_inout:
                yield node, min(t2_inout)
        else:
            # If t1_inout and t2_inout were both empty....
            # P(s) = (N_1 - M_1) x {min (N_2 - M_2)}
            # # if not (t1_inout or t2_inout):
            # as suggested by  [2], incorrect
            # as inferred from [1], correct
            # First we determine the candidate node for g2
            for node in self.g1.vertices:
                if node not in self.core_1:
                    yield node, min(self.g2_nodes - set(self.core_2))
        # For all other cases, we don't have any candidate pairs.

    cdef bint is_isomorphic(self):
        """Returns True if g1 and g2 are isomorphic graphs."""
        # Let's do two very quick checks!
        # QUESTION: Should we call faster_graph_could_be_isomorphic(g1,g2)?
        # For now, graph3 just copy the code.
        # Check global properties
        if len(self.g1.vertices) != len(self.g2.vertices):
            return False
        # Check local properties
        if (
            sorted([len(neighbor) for neighbor in self.g1.adj.values()])
            != sorted([len(neighbor) for neighbor in self.g2.adj.values()])
        ):
            return False
        self.initialize()
        try:
            next(self.match())
            return True
        except StopIteration:
            return False

    def match(self) -> Iterator[Dict[int, Tuple[int]]]:
        """Extends the isomorphism mapping."""
        cdef int g1_node, g2_node
        cdef GMState new_state
        if len(self.core_1) == len(self.g2.vertices):
            # Save the final mapping, otherwise garbage collection deletes it.
            self.mapping = self.core_1.copy()
            # The mapping is complete.
            yield self.mapping
        else:
            for g1_node, g2_node in self.candidate_pairs_iter():
                if self.syntactic_feasibility(g1_node, g2_node):
                    # Recursive call, adding the feasible state.
                    new_state = type(self.state)(self, g1_node, g2_node)
                    yield from self.match()
                    # restore data structures
                    new_state.restore()

    cdef inline bint syntactic_feasibility(self, int g1_node, int g2_node):
        """Returns True if adding (g1_node, g2_node) is syntactically feasible.

        The VF2 algorithm was designed to work with graphs having, at most,
        one edge connecting any two nodes.  This is not the case when
        dealing with an MultiGraphs.

        Basically, when we test the look-ahead rules R_neighbor, we will
        make sure that the number of edges are checked. We also add
        a R_self check to verify that the number of self loops is acceptable.

        Users might be comparing Graph instances with MultiGraph instances.
        So the generic GraphMatcher class must work with MultiGraphs.
        Care must be taken since the value in the innermost dictionary is a
        singlet for Graph instances.  For MultiGraphs, the value in the
        innermost dictionary is a list.
        """

        # ## Test at each step to get a return value as soon as possible.

        # ## Look ahead 0
        # R_self
        # The number of self loops for g1_node must equal the number of
        # self-loops for g2_node. Without this check, we would fail on
        # R_neighbor at the next recursion level. But it is good to prune the
        # search tree now.
        if (
            _is_adjacent(self.g1, g1_node, g1_node)
            != _is_adjacent(self.g2, g2_node, g2_node)
        ):
            return False
        # R_neighbor
        # For each neighbor n' of n in the partial mapping, the corresponding
        # node m' is a neighbor of m, and vice versa. Also, the number of
        # edges must be equal.
        cdef int neighbor
        for neighbor in self.g1.adj[g1_node]:
            if neighbor in self.core_1:
                if self.core_1[neighbor] not in self.g2.adj[g2_node]:
                    return False
                elif (
                        _is_adjacent(self.g1, neighbor, g1_node)
                        != _is_adjacent(self.g2, self.core_1[neighbor], g2_node)
                ):
                    return False
        for neighbor in self.g2.adj[g2_node]:
            if neighbor in self.core_2:
                if self.core_2[neighbor] not in self.g1.adj[g1_node]:
                    return False
                elif (
                        _is_adjacent(self.g1, self.core_2[neighbor], g1_node)
                        != _is_adjacent(self.g2, neighbor, g2_node)
                ):
                    return False

        # ## Look ahead 1
        # R_term_inout
        # The number of neighbors of n that are in T_1^{inout} is equal to the
        # number of neighbors of m that are in T_2^{inout}, and vice versa.
        cdef int num1 = 0
        for neighbor in self.g1.adj[g1_node]:
            if (neighbor in self.inout_1) and (neighbor not in self.core_1):
                num1 += 1
        cdef int num2 = 0
        for neighbor in self.g2.adj[g2_node]:
            if (neighbor in self.inout_2) and (neighbor not in self.core_2):
                num2 += 1
        if num1 != num2:
            return False

        # ## Look ahead 2
        # R_new
        # The number of neighbors of n that are neither in the core_1 nor
        # T_1^{inout} is equal to the number of neighbors of m
        # that are neither in core_2 nor T_2^{inout}.
        num1 = 0
        for neighbor in self.g1.adj[g1_node]:
            if neighbor not in self.inout_1:
                num1 += 1
        num2 = 0
        for neighbor in self.g2.adj[g2_node]:
            if neighbor not in self.inout_2:
                num2 += 1
        return num1 == num2


@cython.final
cdef class GMState:
    """Graph matcher state."""
    cdef GraphMatcher gm
    cdef int g1_node, g2_node, depth

    def __cinit__(
        self,
        GraphMatcher gm,
        int g1_node=-1,
        int g2_node=-1
    ):
        """Initializes GMState object.

        Pass in the GraphMatcher to which this GMState belongs and the
        new node pair that will be added to the GraphMatcher's current
        isomorphism mapping.
        """
        self.gm = gm

        # Initialize the last stored node pair.
        self.g1_node = -1
        self.g2_node = -1
        self.depth = len(gm.core_1)

        if g1_node == -1 or g2_node == -1:
            # Then we reset the class variables
            gm.core_1 = {}
            gm.core_2 = {}
            gm.inout_1 = {}
            gm.inout_2 = {}

        cdef int node
        # Watch out! g1_node == 0 should evaluate to True.
        if g1_node != -1 and g2_node != -1:
            # Add the node pair to the isomorphism mapping.
            gm.core_1[g1_node] = g2_node
            gm.core_2[g2_node] = g1_node

            # Store the node that was added last.
            self.g1_node = g1_node
            self.g2_node = g2_node

            # Now we must update the other two vectors.
            # We will add only if it is not in there already!
            self.depth = len(gm.core_1)

            # First we add the new nodes...
            if g1_node not in gm.inout_1:
                gm.inout_1[g1_node] = self.depth
            if g2_node not in gm.inout_2:
                    gm.inout_2[g2_node] = self.depth

            # Now we add every other node...

            # Updates for T_1^{inout}
            new_nodes = set()
            for node in gm.core_1:
                new_nodes.update([
                    neighbor for neighbor in gm.g1.adj[node]
                    if neighbor not in gm.core_1
                ])
            for node in new_nodes:
                if node not in gm.inout_1:
                    gm.inout_1[node] = self.depth

            # Updates for T_2^{inout}
            new_nodes = set()
            for node in gm.core_2:
                new_nodes.update([
                    neighbor for neighbor in gm.g2.adj[node]
                    if neighbor not in gm.core_2
                ])
            for node in new_nodes:
                if node not in gm.inout_2:
                    gm.inout_2[node] = self.depth

    cdef void restore(self):
        """Deletes the GMState object and restores the class variables."""
        # First we remove the node that was added from the core vectors.
        # Watch out! g1_node == 0 should evaluate to True.
        if self.g1_node != -1 and self.g2_node != -1:
            self.gm.core_1.pop(self.g1_node)
            self.gm.core_2.pop(self.g2_node)

        # Now we revert the other two vectors.
        # Thus, we delete all entries which have this depth level.
        cdef int node
        for vector in (self.gm.inout_1, self.gm.inout_2):
            for node in tuple(vector):
                if vector[node] == self.depth:
                    vector.pop(node)



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Real-coded Genetic Algorithm

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libc.math cimport pow
from numpy import zeros, float64 as f64
from .utility cimport uint, MAX_GEN, rand_v, rand_i, ObjFunc, Algorithm


@cython.final
cdef class RGA(Algorithm):
    """The implementation of Real-coded Genetic Algorithm."""
    cdef double cross, mutate_f, win, delta
    cdef double[:] new_fitness, f_tmp
    cdef double[:, :] new_pool, tmp

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        self.cross = settings['cross']
        self.mutate_f = settings['mutate']
        self.win = settings['win']
        self.delta = settings['delta']
        self.new_fitness = zeros(self.pop_num, dtype=f64)
        self.new_pool = zeros((self.pop_num, self.dim), dtype=f64)
        self.tmp = zeros((3, self.dim), dtype=f64)
        self.f_tmp = zeros(3, dtype=f64)

    cdef inline double bound(self, int i, double v) nogil:
        """If a variable is out of bound, replace it with a random value."""
        if not self.func.ub[i] >= v >= self.func.lb[i]:
            return rand_v(self.func.lb[i], self.func.ub[i])
        return v

    cdef inline void crossover(self) nogil:
        cdef uint i, s
        for i in range(0, self.pop_num - 1, 2):
            if not rand_v() < self.cross:
                continue
            for s in range(self.dim):
                # first baby, half father half mother
                self.tmp[0, s] = 0.5 * (self.pool[i, s] + self.pool[i + 1, s])
                # second baby, three quarters of father and quarter of mother
                self.tmp[1, s] = self.bound(s, 1.5 * self.pool[i, s]
                                            - 0.5 * self.pool[i + 1, s])
                # third baby, quarter of father and three quarters of mother
                self.tmp[2, s] = self.bound(s, -0.5 * self.pool[i, s]
                                            + 1.5 * self.pool[i + 1, s])
            # evaluate new baby
            self.f_tmp[0] = self.func.fitness(self.tmp[0, :])
            self.f_tmp[1] = self.func.fitness(self.tmp[1, :])
            self.f_tmp[2] = self.func.fitness(self.tmp[2, :])
            # bubble sort: smaller -> larger
            if self.f_tmp[0] > self.f_tmp[1]:
                self.f_tmp[0], self.f_tmp[1] = self.f_tmp[1], self.f_tmp[0]
                self.tmp[0], self.tmp[1] = self.tmp[1], self.tmp[0]
            if self.f_tmp[0] > self.f_tmp[2]:
                self.f_tmp[0], self.f_tmp[2] = self.f_tmp[2], self.f_tmp[0]
                self.tmp[0], self.tmp[2] = self.tmp[2], self.tmp[0]
            if self.f_tmp[1] > self.f_tmp[2]:
                self.f_tmp[1], self.f_tmp[2] = self.f_tmp[2], self.f_tmp[1]
                self.tmp[1], self.tmp[2] = self.tmp[2], self.tmp[1]
            # replace first two baby to parent, another one will be
            self.assign_from(i, self.f_tmp[0], self.tmp[0])
            self.assign_from(i + 1, self.f_tmp[1], self.tmp[1])

    cdef inline double get_delta(self, double y) nogil:
        cdef double r
        if self.task == MAX_GEN and self.stop_at > 0:
            r = <double>self.func.gen / self.stop_at
        else:
            r = 1
        return y * rand_v() * pow(1.0 - r, self.delta)

    cdef inline void mutate(self) nogil:
        cdef uint i, s
        for i in range(self.pop_num):
            if not rand_v() < self.mutate_f:
                continue
            s = rand_i(self.dim)
            if rand_v() < 0.5:
                self.pool[i, s] += self.get_delta(self.func.ub[s]
                                                  - self.pool[i, s])
            else:
                self.pool[i, s] -= self.get_delta(self.pool[i, s]
                                                  - self.func.lb[s])
            # Get fitness
            self.fitness[i] = self.func.fitness(self.pool[i, :])
        self.find_best()

    cdef inline void select(self) nogil:
        """roulette wheel selection"""
        cdef uint i, j, k
        for i in range(self.pop_num):
            j = rand_i(self.pop_num)
            k = rand_i(self.pop_num)
            if self.fitness[j] > self.fitness[k] and rand_v() < self.win:
                self.new_fitness[i] = self.fitness[k]
                self.new_pool[i, :] = self.pool[k, :]
            else:
                self.new_fitness[i] = self.fitness[j]
                self.new_pool[i, :] = self.pool[j, :]
        # in this stage, new_chromosome is select finish
        # now replace origin chromosome
        self.fitness[:] = self.new_fitness
        self.pool[:] = self.new_pool
        # select random one chromosome to be best chromosome,
        # make best chromosome still exist
        self.assign_from(rand_i(self.pop_num), self.best_f, self.best)

    cdef inline void generation(self) nogil:
        self.select()
        self.crossover()
        self.mutate()



# -*- coding: utf-8 -*-

from numpy import ndarray, double
from .utility import ObjFunc


class TestObj(ObjFunc[float]):

    def fitness(self, v: ndarray) -> double:
        ...

    def result(self, v: ndarray) -> float:
        ...



# -*- coding: utf-8 -*-

from typing import Tuple, List, Sequence, FrozenSet, Mapping, Optional, Union
from .expression import VPoint, Coord

_Coord = Tuple[float, float]
_Inputs = Mapping[Tuple[int, int], float]
_PointPair = Union[int, Tuple[int, int]]

class SolverSystem:
    def __init__(
        self,
        vpoints: Sequence[VPoint],
        inputs: Optional[Mapping[Tuple[int, int], float]] = None,
        data_dict: Optional[Mapping[_PointPair, Union[Coord, float]]] = None
    ):
        """The expression `vpoints` solver function of BFGS method by giving
        the input pairs `inputs` and link length `data_dict` requirements.

        !!! note
            The format of input pairs:

            + Revolut joints: `{(base, driver): angle}`
            + Slider joints: `{(base, base): offset}`

        The format of `data_dict`:

        + Specific coordinates: Dict\[int, List\[Coord]]
        + Specific link length: Dict\[Tuple\[int, int], float]

        The `data_dict` parameter will reformat its keys into `frozenset` type.
        """
        ...

    def same_points(self, vpoints_: Sequence[VPoint]) -> bool:
        ...

    def show_inputs(self) -> FrozenSet[Tuple[int, int]]:
        ...

    def show_data(self) -> FrozenSet[_PointPair]:
        ...

    def set_inputs(self, inputs: Mapping[Tuple[int, int], float]) -> None:
        ...

    def set_data(self, data_dict: Union[_Inputs, Mapping[int, Coord]]) -> None:
        ...

    def solve(self) -> List[Union[_Coord, Tuple[_Coord, _Coord]]]:
        ...



# -*- coding: utf-8 -*-

"""Optimization utilities."""

__author__ = "Yuan Chang"
__copyright__ = "Copyright (C) 2016-2021"
__license__ = "AGPL"
__email__ = "pyslvs@gmail.com"

from typing import TypedDict, Sequence, Tuple, Dict
from pyslvs.expression import VPoint

_Path = Sequence[Tuple[float, float]]


class FConfig(TypedDict, total=False):
    expression: Sequence[VPoint]
    input: Sequence[Tuple[Tuple[int, int], Sequence[float]]]
    placement: Dict[int, Tuple[float, float, float]]
    target: Dict[int, _Path]
    same: Dict[int, int]
    upper: float
    lower: float
    shape_only: bool


class NConfig(TypedDict, total=False):
    target: _Path
